# Synced from ballistica-master-server.
# EFRO_SYNC_HASH=253867589410814029404123679015679858153
#
# Copyright (c) 2011-2024 Eric Froemling
"""Functionality related to cloud-stored files."""

from __future__ import annotations

from typing import TYPE_CHECKING
import hashlib
import base64

if TYPE_CHECKING:
    from typing import IO

NEW = True

# NOTES TO SELF:
#
# The reason for having this concept of 'cloud-file-ids' instead of
# simply passing around hashes is so we can gracefully change the
# hashing algorithm or otherwise change the format of the ids later. For
# now we are simply tacking on a bit of extra data to a hash to force
# code to use these conversion functions instead of treating these as
# hashes.
#
# IMPORTANT DESIGN CONSIDERATION:
#
# - At first it seemed that it would be harmless to have multiple
#   cloud-file-ids that describe the same data (say, a sha256 id as well
#   as a sha512 id or whatever). However this is NOT ok. This can lead
#   to breakage; for example redundantly sending clients files due to
#   thinking they have different versions based on cloud-file-ids we
#   construct from hashes of their files which come out different than
#   old-format cloud-file-ids we're holding onto even though the
#   underlying data is the same. So it is very important that a given
#   file is represented by exactly ONE file id everywhere.
#
# If we change the cloud-file-id format we need to do the following:
#
# - Update all live servers to understand/allow the new format while
#   still creating everything with the old.
#
# - Iterate through the db and, for everything that holds a
#   cloud-file-id, read the file and then write it to convert to the new
#   id format. Note that this is basically the list of things that are
#   checked in check_if_used().
#
# - Flip live servers to use the new format.
#
# - Do one more db pass to convert any old id forms to new that were
#   created after the first iteration but before we flipped the default.
#
# - Old form CloudFiles should then simply die off due to no references.
#   Though it could be a good idea to iterate through them and
#   explicitly make sure they are unreferenced (and possibly just kill
#   them immediately or push up their expire times).
#
# - As a sanity check, the below functions can then be set to error/warn
#   if old formats are passed through. Nothing should be doing that.

# FURTHER THOUGHTS:
#
# - ACTUALLY - Wouldn't the above problem just work out ok if we did the
#   comparisons using sha256 or whatever the client is dealing with
#   instead of converting to cloud-file-ids? In that case we wouldn't
#   redundantly send the client files, but when they *do* sent us a
#   changed one it would be stored in the new id format.
#
# - Though I guess the problem could still manifest if we ever hash some
#   cloud-file-ids without thinking about it/etc. So maybe it would
#   still be good to stick to the exactly-one-id-per-file rule.
#
# - What if cloud-file-ids were simply flattened/sorted json dicts? We
#   could store useful metadata such as file-size or detected format or
#   image resolution or multiple hashes or whatever (anything purely
#   based on file data itself). Firestore document ids can be up to
#   1,500 bytes so we've got plenty of room.
#
# - We could store cloud-file-id version with the file to make it easy
#   to update old ones if that would be useful. Though I wonder if
#   format would remain unchanged after a while, making that wasted
#   space/indexing.
#
# - Also, since we know exactly where in the db to look to find all
#   references to a given file, we could automate the process of
#   upgrading all file references to newer cloud-file-id versions,
#   making it a mostly painless process.
#
# - Though wondering in what cases an embedded size would be more useful
#   than having to look up the CloudFile document. We could use it to
#   avoid adding size to WorkspaceDirEntries. It could also just be
#   useful at a glance when debugging build file listings/etc. Perhaps
#   that would be reasonable motivation for a file-format value too.
#
# - Further stuff like image resolution or format seems like it might be
#   overkill. We could implement this on CloudFile in a 'calculate as
#   requested' fashion.
#
# - What about a CRC32 checksum? This could be 6 base64 chars or 8 hex
#   chars (or 4 bytes if sending binary). Perhaps that could be useful
#   for a quick first pass for syncing; if crc differs then we'd know we
#   need to send the file; otherwise we check full hashes. Though I
#   think that optimizes the wrong case; we want to send as little data
#   as possible when things *don't* change.
#
# - Or at the end of the day maybe what we have is ok; we store size
#   alongside when it matters, path/type when it matters, etc. Since
#   we'd need to do a full conversion for any changes anyway to keep the
#   one-id-per-file rule intact, maybe just worrying about that later if
#   we ever switch hashing types is the best option.


def _id_for_digest_and_size(digest: bytes, size: int) -> str:

    # Let's go with urlsafe base64 with the padding '='s stripped. This
    # gets us pretty compact size without too much need for escaping.
    hashstr = base64.urlsafe_b64encode(digest).decode().replace('=', '')
    return f'F.{size}.{hashstr}'


def cloud_file_id_for_data(data: bytes) -> str:
    """Given complete file data, return its cloud-file-id.

    This id may incorporate a hash or other metadata, but do not make
    any assumption about that, as the format may change over time. Treat
    this as a black box id value.
    """
    sha = hashlib.sha256()
    sha.update(data)

    if NEW:
        return _id_for_digest_and_size(sha.digest(), len(data))

    return 'F' + sha.hexdigest()


def cloud_file_id_for_file(file: IO[bytes]) -> str:
    """Given a file-like-object, return its cloud-file-id.

    Be aware that this will read the entire file.

    This id may incorporate a hash or other metadata, but do not make
    any assumption about that, as the format may change over time. Treat
    this as a black box id value.
    """
    sha = hashlib.sha256()
    size = 0
    for block in iter(lambda: file.read(8192), b''):
        size += len(block)
        sha.update(block)

    if NEW:
        return _id_for_digest_and_size(sha.digest(), size)

    return 'F' + sha.hexdigest()


def cloud_file_id_for_path(path: str) -> str:
    """Given a path to an existing file, return its cloud-file-id.

    This id may incorporate a hash or other metadata, but do not make
    any assumption about that, as the format may change over time. Treat
    this as a black box id value.
    """

    with open(path, 'rb') as file:
        return cloud_file_id_for_file(file)


# Need to kill this if we want the option to include metadata with an id
# (since calcing an id will require the full data in that case).
def cloud_file_id_from_sha256_and_size(hash_sha256: str, size: int) -> str:
    """Given a sha256 hex hash of a file, return its file id."""

    if NEW:
        return _id_for_digest_and_size(bytes.fromhex(hash_sha256), size)

    return 'F' + hash_sha256


def sha256_from_cloud_file_id(file_id: str) -> str:
    """Given a cloud file id, return its sha256 hex hash."""

    # Current format is
    # 'F.(size).(urlsafe-b64-sha256-hash-with-equals-stripped)'. The
    # size here can vary so we check it first so we don't mistake it as
    # another format based on size.
    splits = file_id.split('.')
    if file_id.startswith('F.') and len(splits) == 3:
        # Our ids are urlsafe base64 with trailing '='s stripped out;
        # reconstruct full base64, decode, and then spit out hex.
        stripped = splits[2].encode()
        reconstructed = stripped + b'=' * (len(stripped) % 4)
        digest = base64.urlsafe_b64decode(reconstructed)
        return digest.hex()

    # Old form that was nothing but the sha256 value in hex. Can delete this
    # soon.
    if len(file_id) == 64:
        import logging

        logging.exception('Found old format file-id %s', file_id)
        return file_id

    # Previously we tacked on a single 'F' to the beginning of a hex
    # base64 digest to lightly differentiate cloud-file-ids from raw
    # sha256 values.
    if len(file_id) == 65 and file_id[0] == 'F':
        return file_id[1:]

    raise ValueError('Invalid file-id format.')
