# Synced from ballistica-master-server.
# EFRO_SYNC_HASH=98605881956104739791310202783468916419
#
"""Miscellaneous data structures shared between server components."""

from __future__ import annotations

import datetime
from typing import TYPE_CHECKING, Annotated
from dataclasses import dataclass, field

from typing_extensions import override
from efro.util import utc_now
from efro.dataclassio import ioprepped, IOAttrs, IOExtendedData
from efro.log import LogLevel
from baserver.enums import ServerNodeRunMode

if TYPE_CHECKING:
    pass

MESSAGE_SIDECAR_TIMESTAMP_KEY = '_mts'


@ioprepped
@dataclass
class BASNConfig(IOExtendedData):
    """Config options passed to new basn instances.

    Note to self: basn_bootstrap accesses the json data for this
    directly (namely payload_url and lib_root), so make sure to check
    there if changing any keys as type-checking won't catch errors there.
    """

    node_id: str
    fleet: str
    boot_token: str
    master_server_addr: str
    bootstrap_url: str
    payload_url: str
    user: str
    var_root: str
    log_path: str
    lib_root: str
    num_users: int
    run_mode: ServerNodeRunMode

    # TCP port where we host ourself.
    port: int

    # TCP port where we listen for transport connections.
    transport_port: int

    # UDP port range we can host stuff at.
    udp_port_min: int
    udp_port_max: int

    # python version basn runs under
    python_version: str

    # all python versions we want installed
    all_python_versions: list[str]

    # Our extended-data method to mutate legacy data into something
    # we can digest.
    @override
    @classmethod
    def will_input(cls, data: dict) -> None:
        pass

    @classmethod
    def python_bin_path_for_lib_root_and_version(
        cls, lib_root: str, version: str
    ) -> str:
        """Return Python bin path for arbitrary version."""
        return f'{lib_root}/venv/{version}/bin/python{version}'

    def python_bin_path(self) -> str:
        """Python bin path for *our* version."""
        return self.python_bin_path_for_lib_root_and_version(
            lib_root=self.lib_root, version=self.python_version
        )


@ioprepped
@dataclass
class BASNStats:
    """Stats a basn instance sends to bamaster about itself.

    These stats are sent within a few seconds of changing on the basn end
    and are stored with the ServerNode document on the bamaster end, so
    should be mostly up to date and almost always available.
    """

    max_log_level: Annotated[LogLevel, IOAttrs('maxloglv')] = LogLevel.INFO
    max_task_log_level: Annotated[LogLevel, IOAttrs('maxtaskloglv')] = (
        LogLevel.INFO
    )
    running_task_count_inline: Annotated[int | None, IOAttrs('runtaskcnt')] = (
        None
    )
    running_task_count_standalone: Annotated[
        int | None, IOAttrs('runtaskcnts')
    ] = None
    tasks_errored: Annotated[bool, IOAttrs('taskserred')] = False

    run_mode: Annotated[ServerNodeRunMode | None, IOAttrs('run_mode')] = None

    # Mem used for just the basn process (in megabytes).
    basn_memory_used: Annotated[float | None, IOAttrs('bmem')] = None

    # Total mem used by all processes (in megabytes).
    system_memory_used: Annotated[float | None, IOAttrs('smem')] = None

    # Total mem available (in megabytes).
    system_memory_total: Annotated[float | None, IOAttrs('smemt')] = None

    # Total swap mem in use (in megabytes).
    system_swap_used: Annotated[float | None, IOAttrs('swp')] = None

    # Swap mem read/written in the last minute (in megabytes).
    system_swap_io: Annotated[float | None, IOAttrs('swpio')] = None

    # Universal load (NOT in local space).
    load: Annotated[float | None, IOAttrs('l')] = None

    oldest_supported_ballistica_build: Annotated[
        int | None, IOAttrs('oldestbld')
    ] = None

    cpu_per_load_unit: Annotated[float | None, IOAttrs('lcpu')] = None
    mem_per_load_unit: Annotated[float | None, IOAttrs('lmem')] = None

    # How much memory we use idle.
    memory_idle: Annotated[float | None, IOAttrs('memi')] = None

    # Our hard capacity in universal load units.
    hard_capacity: Annotated[float | None, IOAttrs('hcp')] = None

    disk_used: Annotated[int | None, IOAttrs('duse')] = None

    disk_free: Annotated[int | None, IOAttrs('dfree')] = None

    disk_total: Annotated[int | None, IOAttrs('dtotal')] = None

    open_sockets_count: Annotated[int | None, IOAttrs('osc')] = None
    open_files_count: Annotated[int | None, IOAttrs('ofc')] = None


@ioprepped
@dataclass
class BASNStatsSparse:
    """Stats a basn instance sends to bamaster about itself.

    These stats are sent rarely (once per minute or two) and are stored
    in memcache, so may be more out of date or unavailable. Check create_time
    if out-of-date data is an issue.
    """

    # How often these get sent out to master.
    SHIP_INTERVAL_SECONDS = 61.389  # Roughly once per minute.

    create_time: Annotated[datetime.datetime, IOAttrs('t')] = field(
        default_factory=utc_now
    )

    @dataclass
    class TransportClientStats:
        """Stats for an individual transport-client task."""

        # Lists of handle-ids per session on this transport client.
        session_channels: Annotated[dict[str, list[int]], IOAttrs('sc')]

    # The currently connected set of transport clients.
    transport_clients: Annotated[
        dict[int, TransportClientStats], IOAttrs('tc')
    ] = field(default_factory=dict)
