# Synced from ballistica-master-server.
# EFRO_SYNC_HASH=195344908027655466525463627027241838930
#
"""Functionality pertaining to cloud firestore."""

# pylint: disable=too-many-lines
from __future__ import annotations

import os
import copy
import weakref
import datetime
from typing import TYPE_CHECKING, TypeVar, Generic, get_args, assert_never

# This should be the only place pulling in *actual* firestore.
# All actual usage should be going through our wrapper apis so
# our logging/caching/etc. layers work consistently.

# Note: we used to maintain our own typing stubs for firestore, but it
# seems that they are mostly typed these days. However we have to access
# things via 'firestore_v1' instead of the re-exported plain 'firestore'.
# See discussion here:
# https://github.com/googleapis/python-firestore/issues/447

import google.cloud.firestore_v1
import google.cloud.firestore_v1.base_query
import google.cloud.firestore_v1.field_path

import google.api_core.exceptions

# The user may want to catch these when calling transactional funcs
# containing create-only writes. We could create our own wrapper
# exception, but let's just directly expose Google's for now.
# pylint: disable=useless-import-alias, unused-import
from google.api_core.exceptions import AlreadyExists as AlreadyExists


# pylint: enable=useless-import-alias, unused-import

from efro.terminal import Clr
from efro import dataclassio

if TYPE_CHECKING:
    import logging
    from typing import Any, Callable, Iterable, Iterator, Literal

T = TypeVar('T')
U = TypeVar('U')
DocT = TypeVar('DocT', bound='FirestoreDocument')

DEBUG_PRINT = os.environ.get('BASERVER_FIRESTORE_DEBUG') == '1'

DEBUG_PRINT_SIZE_CALC = False

# An optimization to suppress writes in a transaction that appear to
# be the same data as was read.
# Currently this seems to only reduce writes by ~2% so I don't
# think its worth the extra memory and cpu it requires.
SUPPRESS_REDUNDANT_WRITES = False


def transactional(to_wrap: T) -> T:
    """Wrap the firestore transactional decorator."""

    # Actual code
    if not TYPE_CHECKING:
        # The Firestore transactional() decorator intercepts the first arg
        # to the call which it expects to be a Firestore Transaction.
        # Ours does the same except that it expects our own Transaction type.
        # it then translates that to the underlying firestore Transaction
        # which is passed to an actual @transactional-wrapped call which
        # then calls the original call with our original wrapped Transaction.
        # Phew.

        def _unwrap_transaction(transaction, *args, **keywds):
            # This will get called by the user with our Transaction type
            # as its first arg. Translate it to a raw firestore transaction
            # and pass it to an actual @transactional wrapped call.
            assert isinstance(transaction, Transaction)

            @google.cloud.firestore_v1.transactional
            def _fs_transaction_call(fstransaction, *args2, **keywds2):
                # Firestore calls us (possibly repeatedly) with the raw
                # firestore transaction as first arg. We in turn call our
                # original call with our original wrapper Transaction.
                assert transaction.wrapped is fstransaction
                return to_wrap(transaction, *args2, **keywds2)

            return _fs_transaction_call(transaction.wrapped, *args, **keywds)

        return _unwrap_transaction

    # Dummy code
    return to_wrap


class Client:
    """Wraps Firestore Client class."""

    def __init__(self) -> None:
        self.wrapped = google.cloud.firestore_v1.Client()

        # NOTE: if the number of callbacks here starts getting too
        # big, we should add a registration mechanism to add callbacks
        # for particular collection ids, etc.
        self.did_read_standalone_callbacks: list[Callable[[str], None]] = []
        self.did_read_query_callbacks: list[Callable[[str], None]] = []
        self.did_read_transaction_callbacks: list[Callable[[str], None]] = []
        self.did_write_callbacks: list[Callable[[str], None]] = []
        self.did_redundant_write_callbacks: list[Callable[[str], None]] = []
        self.did_delete_callbacks: list[Callable[[str], None]] = []

    def collection(self, *collection_path: str) -> CollectionReference:
        """Wraps collection call."""
        return CollectionReference(
            client=self, wrapped=self.wrapped.collection(*collection_path)
        )

    def collections(self) -> list[CollectionReference]:
        """Return all top level collections."""
        return [
            CollectionReference(client=self, wrapped=r)
            for r in self.wrapped.collections()
        ]

    def transaction(self) -> Transaction:
        """Wrap transaction call."""
        return Transaction(client=self, wrapped=self.wrapped.transaction())

    def batch(self) -> WriteBatch:
        """Wrap batch call."""
        return WriteBatch(client=self, wrapped=self.wrapped.batch())

    def get_all(
        self,
        references: list[DocumentReference],
        field_paths: Iterable[str] | None = None,
        transaction: Transaction | None = None,
    ) -> Iterable[DocumentSnapshot | None]:
        """Wrap get_all call."""

        # NOTE (Feb 2023): Google's get_all() type annotations currently
        #  tell us that a Generator of DocumentSnapshots is returned, but
        # its inline 'yields' docs mention it can give None for nonexistent
        # documents. So we need to be careful to include the None case here.
        # (If we try to use a lambda instead of defining _wrap(), mypy
        # complains and tell us that the None case can never happen).
        def _wrap(
            snapshot: google.cloud.firestore_v1.DocumentSnapshot | None,
        ) -> DocumentSnapshot | None:
            return (
                DocumentSnapshot(client=self, wrapped=snapshot)
                if snapshot is not None
                else None
            )

        return map(
            _wrap,
            self.wrapped.get_all(
                references=[r.wrapped for r in references],
                field_paths=field_paths,
                transaction=(
                    transaction.wrapped if transaction is not None else None
                ),
            ),
        )

    def log_did_read_query(self, path: str) -> None:
        """Log a completed read associated with a query.

        (Note: this includes both transactional and non-transactional).
        """
        if DEBUG_PRINT:
            print(f'{Clr.GRN}RD{Clr.RST} {path}')
        for callback in self.did_read_query_callbacks:
            callback(path)

    def log_did_read_standalone(self, path: str) -> None:
        """Log a completed read not in a transaction or query."""
        if DEBUG_PRINT:
            print(f'{Clr.GRN}RD{Clr.RST} {path}')
        for callback in self.did_read_standalone_callbacks:
            callback(path)

    def log_did_read_transaction(self, path: str) -> None:
        """Log a completed read that was part of a transaction."""
        if DEBUG_PRINT:
            print(f'{Clr.GRN}RD{Clr.RST} {path}')
        for callback in self.did_read_transaction_callbacks:
            callback(path)

    def log_did_write(self, path: str) -> None:
        """Log a completed write.

        Be sure to only call this *AFTER* the db op has completed.
        This may trigger cache invalidations, and calling this before
        the actual db changes could result in old states being re-fetched.
        """
        if DEBUG_PRINT:
            print(f'{Clr.MAG}WR{Clr.RST} {path}')
        for callback in self.did_write_callbacks:
            callback(path)

    def log_did_redundant_write(self, path: str) -> None:
        """Log a completed redundant write.

        Be sure to only call this *AFTER* the db op has completed.
        This may trigger cache invalidations, and calling this before
        the actual db changes could result in old states being re-fetched.
        """
        if DEBUG_PRINT:
            print(f'{Clr.MAG}RWR{Clr.RST} {path}')
        for callback in self.did_redundant_write_callbacks:
            callback(path)

    def log_did_delete(self, path: str) -> None:
        """Log a completed delete.

        Be sure to only call this *AFTER* the db op has completed.
        This may trigger cache invalidations, and calling this before
        the actual db changes could result in old states being re-fetched.
        """
        if DEBUG_PRINT:
            print(f'{Clr.YLW}DL{Clr.RST} {path}')
        for callback in self.did_delete_callbacks:
            callback(path)


class DocumentReference:
    """Wrapper around Firestore DocumentReference.

    This allows us to inject logging/stats/caching functionality.
    """

    def __init__(
        self,
        client: Client,
        wrapped: google.cloud.firestore_v1.DocumentReference,
    ) -> None:
        assert isinstance(wrapped, google.cloud.firestore_v1.DocumentReference)
        self.client = client
        self.wrapped = wrapped
        self._pending_read = False
        self._pending_write = False
        self._pending_delete = False

    def __del__(self) -> None:
        # Push out any logs that have built up.
        # Hmm; technically we could probably log them immediately
        # after they occur, right? Not like transactions...
        if self._pending_read:
            self.client.log_did_read_standalone(self.path)
        if self._pending_write:
            self.client.log_did_write(self.path)
        if self._pending_delete:
            self.client.log_did_delete(self.path)

    def log_pending_read(self, transaction: Transaction | None) -> None:
        """Log a read operation.

        Can be called within a transaction and either before or after
        the actual db call; the actual logging doesn't happen until
        the containing DocumentReference or Transaction gets garbage-collected.
        """
        # Catch bad ids here so they don't pollute our write stats or
        # give us cryptic firestore errors. These can tend to sneak in
        # through mucked-with URL args.
        validate_document_id(self.id)

        # Have the transaction log this after it completes, or log it ourself
        # if there's no transaction.
        if transaction is not None:
            transaction.log_pending_read(self.path)
        else:
            self._pending_read = True

    def log_pending_write(self, transaction: Transaction | None) -> None:
        """Log a write operation.

        Can be called within a transaction and either before or after
        the actual db call; the actual logging doesn't happen until
        the containing DocumentReference or Transaction gets garbage-collected.
        """
        # Catch bad ids here so they don't pollute our write stats or
        # give us cryptic firestore errors. These can tend to sneak in
        # through mucked-with URL args.
        validate_document_id(self.id)

        # Have the transaction log this after it completes, or log it ourself
        # if there's no transaction.
        if transaction is not None:
            transaction.log_pending_write(self.path)
        else:
            self._pending_write = True

    def log_pending_delete(self, transaction: Transaction | None) -> None:
        """Log a delete operation.

        Can be called within a transaction and either before or after
        the actual db call; the actual logging doesn't happen until
        the containing DocumentReference or Transaction gets garbage-collected.
        """
        # Catch bad ids here so they don't pollute our write stats or
        # give us cryptic firestore errors. These can tend to sneak in
        # through mucked-with URL args.
        validate_document_id(self.id)

        # Have the transaction log this after it completes, or log it ourself
        # if there's no transaction.
        if transaction is not None:
            transaction.log_pending_delete(self.path)
        else:
            self._pending_delete = True

    @property
    def id(self) -> str:
        """The id for the docref."""
        out = self.wrapped.id
        assert isinstance(out, str)
        return out

    @property
    def path(self) -> str:
        """The path for the docref."""
        out = self.wrapped.path
        assert isinstance(out, str)
        return out

    @property
    def parent(self) -> CollectionReference:
        """Return the collection this document is in."""
        return CollectionReference(
            client=self.client, wrapped=self.wrapped.parent
        )

    def create(self, document_data: dict[str, Any]) -> None:
        """Create a document."""
        # Catch bad ids here so they don't pollute our write stats or
        # give us cryptic firestore errors. These can tend to sneak in
        # through mucked-with URL args.
        validate_document_id(self.wrapped.id)

        # NOTE: the actual firestore call returns a WriteResult, which
        # is an internal type. This contains an update_time and
        # transform_results. Can expose this if needed.
        self.log_pending_write(transaction=None)
        self.wrapped.create(document_data=document_data)

    def get(self, transaction: Transaction | None = None) -> DocumentSnapshot:
        """Get a document."""
        # Catch bad ids here so they don't pollute our write stats or
        # give us cryptic firestore errors. These can tend to sneak in
        # through mucked-with URL args.
        validate_document_id(self.wrapped.id)

        self.log_pending_read(transaction=transaction)
        return DocumentSnapshot(
            client=self.client,
            wrapped=self.wrapped.get(
                transaction=(
                    transaction.wrapped if transaction is not None else None
                )
            ),
        )

    def set(self, document_data: dict[str, Any], merge: bool = False) -> None:
        """Set document data."""
        # Catch bad ids here so they don't pollute our write stats or
        # give us cryptic firestore errors. These can tend to sneak in
        # through mucked-with URL args.
        validate_document_id(self.wrapped.id)

        # NOTE: the actual firestore call returns a WriteResult, which
        # is an internal type. This contains an update_time and
        # transform_results. Can expose this if needed.
        self.log_pending_write(transaction=None)
        self.wrapped.set(document_data=document_data, merge=merge)

    def update(self, field_updates: dict[str, Any], option: Any = None) -> None:
        """Update data on an existing document.

        Will error if the document is not found.
        """
        # Catch bad ids here so they don't pollute our write stats or
        # give us cryptic firestore errors. These can tend to sneak in
        # through mucked-with URL args.
        validate_document_id(self.wrapped.id)

        # NOTE: the actual firestore call returns a WriteResult, which
        # is an internal type. This contains an update_time and
        # transform_results. Can expose this if needed.
        self.log_pending_write(transaction=None)
        self.wrapped.update(field_updates=field_updates, option=option)

    def delete(self) -> None:
        """Delete a document."""
        # Catch bad ids here so they don't pollute our write stats or
        # give us cryptic firestore errors. These can tend to sneak in
        # through mucked-with URL args.
        validate_document_id(self.wrapped.id)

        # NOTE: the actual firestore call returns a timestamp. Can
        # expose this if needed.
        self.log_pending_delete(transaction=None)
        self.wrapped.delete()


class DocumentSnapshot:
    """Wrapper around Firestore DocumentSnapshot.

    This allows us to inject logging/stats/caching functionality.
    """

    def __init__(
        self,
        client: Client,
        wrapped: google.cloud.firestore_v1.DocumentSnapshot,
    ) -> None:
        assert isinstance(wrapped, google.cloud.firestore_v1.DocumentSnapshot)
        self.client = client
        self.wrapped = wrapped

    @property
    def exists(self) -> bool:
        """Does this snapshot exist?"""
        out = self.wrapped.exists
        assert isinstance(out, bool)
        return out

    @property
    def reference(self) -> DocumentReference:
        """Reference for this snapshot."""
        return DocumentReference(
            client=self.client, wrapped=self.wrapped.reference
        )

    def to_dict(self) -> dict[str, Any] | None:
        """Convert to dict."""
        return self.wrapped.to_dict()

    def to_dict_strong(self) -> dict[str, Any]:
        """Convert to dict; error if None."""
        out = self.to_dict()
        if out is None:
            raise RuntimeError('to_dict gave None (nonexistent document?)')
        return out

    def calculate_storage_size(self) -> int:
        """Calculates how much quota this snapshot costs.

        Based on https://firebase.google.com/docs/firestore/storage-size
        This assumes that any 'nin' dict at the top level is unindexed
        and all other fields are indexed.

        This doesn't take into account composite indexes, and may not
        be 100% accurate, so try to use it in a 'fuzzy' manner.

        Remember max size per document is 1MB (1,048,576 bytes).
        """

        d = self.to_dict_strong()

        # (32 additional bytes per document)
        fullsize = 32

        # Document name size.
        doc_name_size = 0
        fullpath = self.reference.path
        fullpathsplits = fullpath.split('/')

        # Note: currently assuming we don't have a parent collection
        # for calculations; need to update if we ever do.
        if len(fullpathsplits) > 2:
            raise RuntimeError('FIXME: need to update for nested collections.')
        for component in fullpathsplits:
            doc_name_size += len(component.encode()) + 1
        doc_name_size += 16

        if DEBUG_PRINT_SIZE_CALC:
            print(f'DOC NAME SIZE IS {doc_name_size}')

        fullsize += doc_name_size

        fullsize += _calc_dict_size([], d)

        # Calc single-field index sizes for all top-level fields
        # besides 'nin'.
        for key, val in d.items():
            if key != 'nin':
                index_size = 32  # Index machinery.
                index_size += doc_name_size
                index_size += len(key.encode()) + 1
                index_size += _calc_value_size([f'(index:{key})'], val)
                if DEBUG_PRINT_SIZE_CALC:
                    print(f'FULL-INDEX-SIZE FOR {key} is {index_size}')
                fullsize += index_size

        if DEBUG_PRINT_SIZE_CALC:
            print('FINAL TOTAL SIZE IS', fullsize)
        return fullsize

    def get(self, field_path: str) -> Any:
        """Get a value from a document."""
        return self.wrapped.get(field_path=field_path)


def _calc_dict_size(path: list[str], vals: dict) -> int:
    total_size = 32  # Dict guts.
    for key, val in vals.items():
        assert isinstance(key, str)
        keysize = len(key.encode()) + 1
        valsize = _calc_value_size(path + [key], val)
        total_size += keysize + valsize

    return total_size


def _calc_list_size(path: list[str], vals: list) -> int:
    total_size = 0
    for i, val in enumerate(vals):
        total_size += _calc_value_size(path + [str(i)], val)
    return total_size


def _calc_value_size(path: list[str], val: Any) -> int:
    out: int
    if isinstance(val, str):
        out = len(val.encode()) + 1
    elif isinstance(val, (datetime.datetime, int, float)):
        out = 8
    elif val is None or isinstance(val, bool):
        out = 1
    elif isinstance(val, bytes):
        out = len(val)
    elif isinstance(val, dict):
        out = _calc_dict_size(path, val)
    elif isinstance(val, list):
        out = _calc_list_size(path, val)
    else:
        if DEBUG_PRINT_SIZE_CALC:
            print('FIXME CALC SIZE FOR', type(val))
        out = 0

    if DEBUG_PRINT_SIZE_CALC:
        print('FOR', '/'.join(path), 'GOT SIZE', out)
    return out


class FieldPath:
    """Wrapper around Firestore FieldPath."""

    @staticmethod
    def document_id() -> str:
        """Special key for querying by document id."""
        cls = google.cloud.firestore_v1.field_path.FieldPath
        val = cls.document_id()  # type: ignore
        assert isinstance(val, str)
        return val


class CollectionReference:
    """Wrapper around Firestore CollectionReference."""

    def __init__(
        self,
        client: Client,
        wrapped: google.cloud.firestore_v1.CollectionReference,
    ) -> None:
        assert isinstance(
            wrapped, google.cloud.firestore_v1.CollectionReference
        )
        self.client = client
        self.wrapped = wrapped

    @property
    def id(self) -> str:
        """Last path component of the collection."""
        val = self.wrapped.id
        assert isinstance(val, str)
        return val

    def document(self, document_id: str | None = None) -> DocumentReference:
        """Get a document reference in a collection."""
        return DocumentReference(
            client=self.client,
            wrapped=self.wrapped.document(document_id=document_id),
        )

    def add(
        self, document_data: dict, document_id: str | None = None
    ) -> tuple[Any, Any]:
        """Add to a collection."""

        update_time, out_ref = self.wrapped.add(
            document_data=document_data, document_id=document_id
        )
        assert isinstance(out_ref, google.cloud.firestore_v1.DocumentReference)
        self.client.log_did_write(out_ref.path)
        return update_time, out_ref

    def get_count(self) -> int:
        """(slightly simplified) wrapper around aggregation query.

        Aggregation queries charge one document query per 1000 documents
        counted.

        NOTE: A charge of one document read applies even if there are no
              results, so if it is likely that queries will return no
              documents then it can be cheaper to use regular queries.
        """
        count_query = self.wrapped.count()  # type: ignore
        result = count_query.get()
        assert isinstance(result, list)
        assert len(result) == 1
        assert isinstance(result[0], list)
        assert len(result[0]) == 1
        count = result[0][0].value
        assert isinstance(count, int), f'expected int value; got {type(count)}'
        return count

    def where(self, field_path: str, op_string: str, value: Any) -> Query:
        """Query start."""

        # Unwrap some value types.
        if isinstance(value, DocumentReference):
            value = value.wrapped

        filt = google.cloud.firestore_v1.base_query.FieldFilter(
            field_path=field_path, op_string=op_string, value=value
        )  # type: ignore
        return Query(
            client=self.client,
            wrapped=self.wrapped.where(filter=filt),
            # wrapped=self.wrapped.where(
            #     field_path=field_path, op_string=op_string, value=value
            # ),
        )

    def start_after(self, snapshot: DocumentSnapshot | dict) -> Query:
        """Start after a snapshot."""
        return Query(
            client=self.client,
            wrapped=self.wrapped.start_after(
                snapshot if isinstance(snapshot, dict) else snapshot.wrapped
            ),
        )

    def select(self, field_paths: Iterable[str]) -> Query:
        """Include specific fields."""
        return Query(
            client=self.client,
            wrapped=self.wrapped.select(field_paths=field_paths),
        )

    def stream(
        self, transaction: Transaction | None = None
    ) -> Iterator[DocumentSnapshot]:
        """Stream resulting documents."""

        def _wrap(
            gsnap: google.cloud.firestore_v1.DocumentSnapshot,
        ) -> DocumentSnapshot:
            self.client.log_did_read_query(gsnap.reference.path)
            return DocumentSnapshot(client=self.client, wrapped=gsnap)

        return map(
            _wrap,
            self.wrapped.stream(
                transaction=transaction.wrapped if transaction else None
            ),
        )

    def order_by(
        self,
        field_path: str,
        direction: Literal['ASCENDING', 'DESCENDING'] = 'ASCENDING',
    ) -> Query:
        """Set order."""
        return Query(
            client=self.client,
            wrapped=self.wrapped.order_by(
                field_path=field_path, direction=direction
            ),
        )

    def limit(self, count: int) -> Query:
        """Set limit."""
        return Query(
            client=self.client, wrapped=self.wrapped.limit(count=count)
        )


class Query:
    """Wrapper around Firestore Query."""

    def __init__(
        self,
        client: Client,
        wrapped: google.cloud.firestore_v1.base_query.BaseQuery,
    ) -> None:
        assert isinstance(wrapped, google.cloud.firestore_v1.Query)
        self.client = client
        self.wrapped = wrapped

    def stream(
        self, transaction: Transaction | None = None
    ) -> Iterator[DocumentSnapshot]:
        """Wrapped stream call."""
        from google.api_core import retry

        def _wrap(
            gsnap: google.cloud.firestore_v1.DocumentSnapshot,
        ) -> DocumentSnapshot:
            self.client.log_did_read_query(gsnap.reference.path)
            return DocumentSnapshot(client=self.client, wrapped=gsnap)

        return map(
            _wrap,
            self.wrapped.stream(
                transaction=transaction.wrapped if transaction else None,
                # FIXME - Temp workaround for bug:
                # https://github.com/googleapis/python-firestore/issues/596
                retry=retry.Retry(deadline=60),
            ),
        )

    def get_count(self) -> int:
        """(slightly simplified) wrapper around aggregation query."""
        # Note (Dec '23): type annotation for this says it returns a type
        # but it returns an instance, so need to use Any here.
        count_query: Any = self.wrapped.count()
        result = count_query.get()
        assert isinstance(result, list)
        assert len(result) == 1
        assert isinstance(result[0], list)
        assert len(result[0]) == 1
        count = result[0][0].value
        # Strangely am getting floats here for value 0; not ints.
        assert isinstance(
            count, (int, float)
        ), f'expected int/float value; got {type(count)}'
        return int(count)

    def where(self, field_path: str, op_string: str, value: Any) -> Query:
        """Wrap where call."""

        # Unwrap some value types.
        if isinstance(value, DocumentReference):
            value = value.wrapped

        filt = google.cloud.firestore_v1.base_query.FieldFilter(
            field_path=field_path, op_string=op_string, value=value
        )  # type: ignore
        return Query(
            client=self.client,
            wrapped=self.wrapped.where(filter=filt),
            # wrapped=self.wrapped.where(
            #     field_path=field_path, op_string=op_string, value=value
            # ),
        )

    def start_after(self, snapshot: DocumentSnapshot | dict) -> Query:
        """Start after a snapshot."""
        return Query(
            client=self.client,
            wrapped=self.wrapped.start_after(
                snapshot if isinstance(snapshot, dict) else snapshot.wrapped
            ),
        )

    def select(self, field_paths: Iterable[str]) -> Query:
        """Wrap select call."""
        return Query(
            client=self.client,
            wrapped=self.wrapped.select(field_paths=field_paths),
        )

    def order_by(
        self,
        field_path: str,
        direction: Literal['ASCENDING', 'DESCENDING'] = 'ASCENDING',
    ) -> Query:
        """Wrap order_by call."""
        return Query(
            client=self.client,
            wrapped=self.wrapped.order_by(
                field_path=field_path, direction=direction
            ),
        )

    def limit(self, count: int) -> Query:
        """Wrap limit call."""
        return Query(
            client=self.client, wrapped=self.wrapped.limit(count=count)
        )


class WriteBatch:
    """Wrapper for Firestore WriteBatch."""

    def __init__(
        self, client: Client, wrapped: google.cloud.firestore_v1.WriteBatch
    ) -> None:
        assert isinstance(wrapped, google.cloud.firestore_v1.WriteBatch)
        self.client = client
        self.wrapped = wrapped

        self._pending_reads: set[str] | None = None
        self._pending_writes: set[str] | None = None
        self._pending_deletes: set[str] | None = None

    def __del__(self) -> None:
        # Push out any logs that have built up.
        if self._pending_reads is not None:
            for path in self._pending_reads:
                self.client.log_did_read_standalone(path)
        if self._pending_writes is not None:
            for path in self._pending_writes:
                self.client.log_did_write(path)
        if self._pending_deletes is not None:
            for path in self._pending_deletes:
                self.client.log_did_delete(path)

    def create(
        self, reference: DocumentReference, document_data: dict[str, Any]
    ) -> None:
        """Wrap create call."""
        # Catch bad ids here so they don't pollute our write stats or
        # give us cryptic firestore errors. These can tend to sneak in
        # through mucked-with URL args.
        validate_document_id(reference.id)

        self.log_pending_write(reference.path)
        self.wrapped.create(
            reference=reference.wrapped, document_data=document_data
        )

    def set(
        self,
        reference: DocumentReference,
        document_data: dict[str, Any],
        merge: bool = False,
    ) -> None:
        """Wrap set call."""
        # Catch bad ids here so they don't pollute our write stats or
        # give us cryptic firestore errors. These can tend to sneak in
        # through mucked-with URL args.
        validate_document_id(reference.id)

        self.log_pending_write(reference.path)
        self.wrapped.set(
            reference=reference.wrapped,
            document_data=document_data,
            merge=merge,
        )

    def update(
        self,
        reference: DocumentReference,
        field_updates: dict[str, Any],
        option: Any = None,
    ) -> None:
        """Update data on an existing document.

        Will error if the document does not exist.
        """
        # Catch bad ids here so they don't pollute our write stats or
        # give us cryptic firestore errors. These can tend to sneak in
        # through mucked-with URL args.
        validate_document_id(reference.id)

        self.log_pending_write(reference.path)
        self.wrapped.update(
            reference=reference.wrapped,
            field_updates=field_updates,
            option=option,
        )

    def delete(self, reference: DocumentReference, option: Any = None) -> None:
        """Wrap delete call."""
        # Catch bad ids here so they don't pollute our write stats or
        # give us cryptic firestore errors. These can tend to sneak in
        # through mucked-with URL args.
        validate_document_id(reference.id)

        self.log_pending_delete(reference.path)
        self.wrapped.delete(reference=reference.wrapped, option=option)

    def commit(self) -> None:
        """Wrap commit call."""
        # NOTE: The underlying call returns list[WriteResult]
        # we may want to expose that at some point.
        self.wrapped.commit()

    def log_pending_read(self, path: str) -> None:
        """Log a read operation.

        Note that the actual logging doesn't happen until this Transaction
        instance is garbage-collected.
        """
        if self._pending_reads is None:
            self._pending_reads = set()
        self._pending_reads.add(path)

    def log_pending_write(self, path: str) -> None:
        """Log a write operation.

        Note that the actual logging doesn't happen until this Transaction
        instance is garbage-collected.
        """
        if self._pending_writes is None:
            self._pending_writes = set()
        self._pending_writes.add(path)

    def log_pending_delete(self, path: str) -> None:
        """Log a delete operation.

        Note that the actual logging doesn't happen until this Transaction
        instance is garbage-collected.
        """
        if self._pending_deletes is None:
            self._pending_deletes = set()
        self._pending_deletes.add(path)


class Transaction:
    """Wrapper around FireStore Transaction."""

    def __init__(
        self, client: Client, wrapped: google.cloud.firestore_v1.Transaction
    ) -> None:
        assert isinstance(wrapped, google.cloud.firestore_v1.Transaction)
        self.client = client
        self.wrapped = wrapped

        # Note: technically, transactions can be retried repeatedly
        # and different reads/writes can happen in each cycle. However we
        # accumulate all reads/writes/deletes from all cycles so our logging
        # might include some false positives.
        self._pending_reads: set[str] | None = None
        self._pending_writes: set[str] | None = None
        self._pending_redundant_writes: set[str] | None = None
        self._pending_deletes: set[str] | None = None

    def __del__(self) -> None:
        # Push out any logs that have built up.
        if self._pending_reads is not None:
            for path in self._pending_reads:
                self.client.log_did_read_transaction(path)
        if self._pending_writes is not None:
            for path in self._pending_writes:
                self.client.log_did_write(path)
        if self._pending_redundant_writes is not None:
            for path in self._pending_redundant_writes:
                self.client.log_did_redundant_write(path)
        if self._pending_deletes is not None:
            for path in self._pending_deletes:
                self.client.log_did_delete(path)

    def create(
        self, reference: DocumentReference, document_data: dict[str, Any]
    ) -> None:
        """Wrap create call."""
        self.log_pending_write(reference.path)
        self.wrapped.create(
            reference=reference.wrapped, document_data=document_data
        )

    def set(
        self,
        reference: DocumentReference,
        document_data: dict[str, Any],
        merge: bool = False,
    ) -> None:
        """Wrap set call."""
        self.log_pending_write(reference.path)
        self.wrapped.set(
            reference=reference.wrapped,
            document_data=document_data,
            merge=merge,
        )

    def redundant_write(self, reference: DocumentReference) -> None:
        """Wrap a set that we optimized out."""
        self.log_pending_redundant_write(reference.path)

    def update(
        self,
        reference: DocumentReference,
        field_updates: dict[str, Any],
        option: Any = None,
    ) -> None:
        """Update data on an existing document.

        Will error if the document does not exist.
        """
        self.log_pending_write(reference.path)
        self.wrapped.update(
            reference=reference.wrapped,
            field_updates=field_updates,
            option=option,
        )

    def delete(self, reference: DocumentReference, option: Any = None) -> None:
        """Wrap delete call."""
        self.log_pending_delete(reference.path)
        self.wrapped.delete(reference=reference.wrapped, option=option)

    def log_pending_read(self, path: str) -> None:
        """Log a read operation.

        Note that the actual logging doesn't happen until this Transaction
        instance is garbage-collected.
        """
        if self._pending_reads is None:
            self._pending_reads = set()
        self._pending_reads.add(path)

    def log_pending_write(self, path: str) -> None:
        """Log a write operation.

        Note that the actual logging doesn't happen until this Transaction
        instance is garbage-collected.
        """
        if self._pending_writes is None:
            self._pending_writes = set()
        self._pending_writes.add(path)

    def log_pending_redundant_write(self, path: str) -> None:
        """Log a redundant write operation.

        Note that the actual logging doesn't happen until this Transaction
        instance is garbage-collected.
        """
        if self._pending_redundant_writes is None:
            self._pending_redundant_writes = set()
        self._pending_redundant_writes.add(path)

    def log_pending_delete(self, path: str) -> None:
        """Log a delete operation.

        Note that the actual logging doesn't happen until this Transaction
        instance is garbage-collected.
        """
        if self._pending_deletes is None:
            self._pending_deletes = set()
        self._pending_deletes.add(path)


class InvalidIdError(Exception):
    """Thrown when an invalid id is found.

    Message contains a user-readable description of what makes the
    id invalid; this can often be presented to the user if user-provided
    data is at fault.
    """


def batch_delete(client: Client, refs: list[DocumentReference]) -> None:
    """Delete multiple references efficiently using batch writes."""

    # Delete in batches until the list is empty.
    # Firestore batches support up to 500 writes.
    while refs:
        count = 0
        batch = client.batch()
        while count < 500 and refs:
            batch.delete(refs.pop())
            count += 1
        batch.commit()


def batch_delete_older_than(
    client: Client,
    collection_path: str,
    time_field: str,
    cutoff_time: datetime.datetime,
    logger: logging.Logger,
    logextra: dict,
    classname: str,
) -> int:
    """Delete all documents in a collection with a field older than given.

    Returns the number deleted.
    """

    batchsize = 500
    count = 0
    while True:
        snapshots = (
            client.collection(collection_path)
            .where(time_field, '<', cutoff_time)
            .limit(batchsize)
            .stream()
        )
        drefs = [s.reference for s in snapshots]
        count += len(drefs)
        if drefs:
            batch_delete(client, drefs)
            logger.info(
                '%s delete tally at %d...', classname, count, extra=logextra
            )
        else:
            break
    if count:
        logger.info('Deleted %d expired %s documents.', count, classname)
    return count


class FirestoreDocument(Generic[T]):
    """Maps a dataclass type to a firestore document."""

    class NotFoundError(Exception):
        """Exception thrown when attempting to fetch a nonexistent document."""

    reference: DocumentReference
    _snapshot: DocumentSnapshot | None
    _data: T | None

    # IMPORTANT: If overriding the constructor in a subclass, do not
    # change the __init__ signature; we call it with this signature when
    # making copies via get_again() or copy().

    # Subclasses should define factory classmethods such as
    # get_by_address() to return fully-constructed instances of
    # themselves; this sort of logic should not be defined in __init__().

    def __init__(
        self,
        source: DocumentReference | DocumentSnapshot | CollectionReference,
        explicit_data_type: type[T] | None = None,
    ):
        """Create an instance from a reference, snapshot, or collection.

        If a CollectionReference is passed, a random id will be generated
        within it.
        """

        # Normally we pull data type from the class's Generic param,
        # but in certain cases that is not possible and it must be
        # passed in instead. Note however that cls.get_data_type() will fail
        # in those cases.
        self._data_type: type[T] = (
            explicit_data_type
            if explicit_data_type is not None
            else self.get_data_type()
        )

        # A useful safety switch when documents are intended to be read-only.
        self.readonly = False

        # When set to True, put() will use Firestore create() calls instead
        # of set(), meaning that the presence of an existing document will
        # cause failure.
        self.createonly = False

        # Optimization: if this document's data was fetched as part of
        # a transaction and put() gets called with the same values as part
        # of the same transaction, we can skip the write.
        self._data_from_transaction_get: (
            tuple[weakref.ref[Transaction], Any] | None
        ) = None

        if isinstance(source, DocumentSnapshot):
            self._snapshot = source
            self.reference = self._snapshot.reference
            self._update_data_from_snapshot()
        elif isinstance(source, DocumentReference):
            self.reference = source
            self._snapshot = None
            self._data = None
        elif isinstance(source, CollectionReference):
            # Gives a random document id within the collection.
            self._snapshot = None
            self._data = None
            self.reference = source.document()
        else:
            assert_never(source)

    def get_again(
        self: DocT, transaction: Transaction | None, allow_create: bool
    ) -> DocT:
        """Return a fresh document instance pulled from our current ref

        If allow_create is False and the document is not found in the
        database, a NotFoundError will be thrown.

        If allow_create is True and the document is not found in the
        database, the new document's data is set to a copy of ours
        (which may itself be previously-pulled data or fallback
        defaults).

        If we have no data, a ValueError is raised.

        This method, when combined with steal_data(), can be useful for
        updating a document 'in-place' with the results of a transaction;
        get_again() gets called as part of the transaction and the results
        are returned and sucked back into the caller via steal_data().
        """
        cls = type(self)
        out = cls(self.reference)
        out.get_data(
            transaction=transaction,
            default_factory=(
                self._copy_data_for_get_again if allow_create else None
            ),
        )
        return out

    def copy(self: DocT, readonly: bool = True) -> DocT:
        """Create a copy of this document wrapper.

        The copy will contain an independent copy of our data and its
        reference will point to the same place. This can be useful when
        fetched documents need to be read or modified safely from multiple
        threads.
        """
        cls = type(self)
        out = cls(self.reference)
        out.data = copy.deepcopy(self.data)
        out.readonly = readonly
        return out

    @property
    def snapshot(self) -> DocumentSnapshot:
        """Access the snapshot from the last get_data() or from creation."""
        if self._snapshot is None:
            raise RuntimeError('No snapshot found; perform a get first.')
        return self._snapshot

    @property
    def data(self) -> T:
        """Access data; get() or create() must have been called prior."""
        if self._data is None:
            raise RuntimeError('Data is not set.')
        return self._data

    @data.setter
    def data(self, value: T) -> None:
        """Set fresh data."""
        assert isinstance(value, self._data_type)
        self._data = value
        self._snapshot = None

    def get_data(
        self,
        transaction: Transaction | None,
        default_factory: Callable[[], T] | None = None,
    ) -> None:
        """Fetch data from the database.

        If the document does not exist in the database and default_factory
        is passed, default data will be set from it. Otherwise a
        NotFoundError will be thrown. The value of self.snapshot.exists
        can be used to detect whether a database document was found.
        """
        self._data_from_transaction_get = None  # reset this
        self._snapshot = self.reference.get(transaction=transaction)
        if self._snapshot.exists:
            self._update_data_from_snapshot(transaction)
        else:
            if default_factory is None:
                raise self.NotFoundError(
                    f'Firestore document for type {self._data_type} with'
                    f' id {self.reference.id} not found.'
                    f' Provide default_factory to allow creation.'
                )
            default = default_factory()
            assert isinstance(default, self._data_type)
            self._data = default

    def steal_data(self: DocT, other: DocT) -> None:
        """Steal data from another document.

        The other wrapper will be left in an invalid state.
        """
        if type(other) is not type(self):
            raise TypeError(
                f'Type mismatch; cannot steal data.'
                f' Expected {type(self)}; got {type(other)}'
            )
        # pylint: disable=protected-access
        if other._data is None:
            raise ValueError('Target has no data to steal.')
        self._data = other._data
        other._data = None
        self._snapshot = None

    def put(
        self, transaction: Transaction | None, batch: WriteBatch | None = None
    ) -> None:
        """Write data to the db.

        If a Transaction or WriteBatch is provided, the put
        will be made as a part of that operation.
        """

        if self.readonly:
            raise RuntimeError('Document is read-only')
        if transaction is not None and batch is not None:
            raise RuntimeError("Can't pass both transaction and batch.")
        payload = dataclassio.dataclass_to_dict(
            self.data, codec=dataclassio.Codec.FIRESTORE
        )
        if batch is not None:
            if self.createonly:
                batch.create(self.reference, payload)
            else:
                batch.set(self.reference, payload)
        elif transaction is not None:
            if self.createonly:
                transaction.create(self.reference, payload)
            else:
                # Optimization: if we're setting data identical to what
                # we previously fetched as part of the same transaction,
                # we can skip the write.
                if (
                    SUPPRESS_REDUNDANT_WRITES
                    and self._data_from_transaction_get is not None
                    and self._data_from_transaction_get[0]() is transaction
                    and self._data_from_transaction_get[1] == payload
                ):
                    transaction.redundant_write(self.reference)
                else:
                    transaction.set(self.reference, payload)
        else:
            if self.createonly:
                self.reference.create(payload)
            else:
                self.reference.set(payload)

    def put_fields(
        self, fields: list[str], transaction: Transaction | None
    ) -> None:
        """Push specific fields on this document to the db.

        Fields should be a list of field paths as obtainable from
        self.fieldpaths().

        This will error if the document does not exist.
        """
        from google.cloud.firestore_v1 import DELETE_FIELD

        if self.readonly:
            raise RuntimeError('Document is read-only')

        if self.createonly:
            raise RuntimeError(
                'put_fields() cannot be used with createonly=True'
            )

        fields_out: dict[str, Any] = {}

        assert isinstance(fields, list)
        payload = dataclassio.dataclass_to_dict(
            self.data, codec=dataclassio.Codec.FIRESTORE
        )
        for field in fields:
            path = field.split('.')
            if not path:
                raise ValueError(
                    'Can\'t pass field root to update;'
                    ' just use put() for that.'
                )

            # Pass either data or a delete-field sentinel to the db
            # for each field depending on whether we have a value for
            # it in our dataclassio output.
            field_data = payload
            have_field_data = True
            for name in path:
                if name in field_data:
                    field_data = field_data[name]
                else:
                    have_field_data = False
                    break
            if have_field_data:
                fields_out[field] = field_data
            else:
                fields_out[field] = DELETE_FIELD

        if transaction is not None:
            transaction.update(self.reference, fields_out)
        else:
            self.reference.update(fields_out)

    @classmethod
    def fieldpath(cls, obj_cb: Callable[[T], Any]) -> str:
        """Get a type-checked storage path for some field under our data.

        example:
           print(self.fieldpath(lambda data: data.foo.bar)) -> 'foo.bar'

        This is basically a convenience call to plug our datatype into
        dataclassio.FieldStoragePathCapture().path().
        """
        return dataclassio.DataclassFieldLookup(cls.get_data_type()).path(
            obj_cb
        )

    @classmethod
    def fieldpaths(cls, obj_cb: Callable[[T], list[Any]]) -> list[str]:
        """Same as fieldpath() but fetches multiple paths at once."""
        return dataclassio.DataclassFieldLookup(cls.get_data_type()).paths(
            obj_cb
        )

    @classmethod
    def get_data_type(cls) -> type[T]:
        """Return the data type we contain."""
        try:
            out = get_args(cls.__orig_bases__[0])[0]  # type: ignore
        except Exception as exc:
            raise TypeError(
                f'Unable to determine data type'
                f' for {cls}; did you pass a type'
                f' as the generic arg?'
            ) from exc
        assert isinstance(out, type)
        return out

    def _update_data_from_snapshot(
        self, transaction: Transaction | None = None
    ) -> None:
        assert self._snapshot is not None
        rawdata = self._snapshot.to_dict_strong()
        self._data = dataclassio.dataclass_from_dict(
            self._data_type, rawdata, codec=dataclassio.Codec.FIRESTORE
        )

        # Take note if we fetched this as part of a transaction
        # (so we can skip writes later in the same transaction if the data
        # is unchanged)
        if SUPPRESS_REDUNDANT_WRITES and transaction is not None:
            self._data_from_transaction_get = (
                weakref.ref(transaction),
                copy.deepcopy(rawdata),
            )

    def _copy_data_for_get_again(self) -> T:
        if self._data is None:
            raise ValueError(
                'Document was not found in the db and local'
                ' data was not set.'
            )
        return copy.deepcopy(self._data)


def is_routine_db_error(exc: Exception) -> bool:
    """Return whether an error seems to be one routinely encountered in db ops.

    These sorts of errors are expected from time to time and should be
    gracefully handled whenever possible instead of making a huge fuss
    about it. Otherwise our error logs get quite spammy when things are busy.
    Be sure to at least log these subtly or whatnot, however, as lots of them
    can be indicative of problematic hot-spots.
    """

    if isinstance(exc, google.api_core.exceptions.InternalServerError):
        return True

    # There often doesn't seem to be a clean type of exception to look for
    # or whatnot so we do the best we can by looking at error messages/etc.
    excstr = str(exc)

    if any(
        s in excstr
        for s in [
            'Too much contention on these documents',
            'The referenced transaction has expired or is no longer valid',
            'Failed to commit transaction',
            'The datastore operation timed out',
            'Aborted due to cross-transaction contention',
        ]
    ):
        return True
    return False


def handle_routine_db_error(
    exc: Exception,
    where: str,
    logger: logging.Logger,
    extra: dict | None = None,
) -> bool:
    """Log, warn, or otherwise take note of contention errors.

    Return whether the error was a routine db error (and thus handled).
    Generally this is called as a last-resort when handling exceptions
    and then the exception is re-raised if this returns False.
    """
    if is_routine_db_error(exc):
        logger.warning('Routine db error at %s.', where, extra=extra)
        return True
    return False


def is_valid_auto_document_id(idstring: str) -> bool:
    """Return whether a given string seems to be a valid autogenerated id.

    Useful for user-provided ids that can be tampered with; factors out
    crazy strings which would results in various ugly document-key errors in
    db operations.
    """
    # NOTE: Should we simply run validate_document_id here? Any strong
    # reason to treat auto ids differently than others?
    # I suppose this is faster and more likely to catch tampering in cases
    # where the resulting string is technically still a valid db name.

    # Currently ids seem to be 20 alphanumeric chars but let's allow a
    # wider range in case Google changes this under us.
    return idstring.isalnum() and (10 < len(idstring) < 50)


def is_valid_document_id(idstring: str) -> bool:
    """Return whether a string is a valid firestore document id."""
    try:
        validate_document_id(idstring)
    except InvalidIdError:
        return False
    return True


def validate_document_id(idstring: str) -> None:
    """Ensure that a provided id is ok to use as a Firestore document id.

    InvalidIdError is raised on any problems.
    """
    assert isinstance(idstring, str)

    if not idstring:
        raise InvalidIdError('The provided value cannot be empty.')

    # Rules copied from here:
    # https://firebase.google.com/docs/firestore/quotas#limits

    # Google specifies utf8 strings no longer than 1500 bytes;
    # Let's be a bit more conservative just in case encoding methods vary
    # or something.
    utf8str = idstring.encode()
    if len(utf8str) > 1400:
        raise InvalidIdError('The provided value is too long.')

    if '/' in idstring:
        raise InvalidIdError(
            "The provided value contains '/' which is illegal."
        )

    if idstring in ('.', '..'):
        raise InvalidIdError("The provided value cannot equal '.' or '..'")

    if idstring.startswith('__.') and idstring.endswith('__'):
        raise InvalidIdError(
            "The provided value cannot start with '__.' and end with '__'"
        )
