# Synced from ballistica-master-server.
# EFRO_SYNC_HASH=322648153750580320719907236627801546272
#
"""Rendering SVG imagery."""

from __future__ import annotations

import html
import uuid
import math
from enum import Enum
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from typing_extensions import override

if TYPE_CHECKING:
    from typing import Sequence


class SVG:
    """An SVG context; results in a single SVG html object."""

    def __init__(
        self,
        width: float = 800.0,
        height: float = 250.0,
        elems: list[Drawable] | None = None,
        show_bounds: bool = False,
    ) -> None:
        self.width = width
        self.height = height
        self.elems = [] if elems is None else elems
        self.show_bounds = show_bounds

    def draw(self) -> str:
        """Convert all drawables to html code."""
        out = f'<svg width="{self.width:.0f}" height="{self.height:.0f}">\n'

        # Draw overall SVG bounds.
        if self.show_bounds:
            out += (
                '<rect width="100%" height="100%"'
                ' fill="none" stroke="rgba(127, 127, 127, 0.5)"'
                ' stroke-width="2"/>\n'
            )

        # Draw all elements.
        for elem in self.elems:
            elem_out = elem.draw(self)
            if elem_out:
                out += f'\n{elem_out}'

        # Fallback text.
        out += 'Sorry, your browser does not support inline SVG.\n' '</svg>'
        return out


class Drawable:
    """An object that can draw itself as part of an SVG."""

    def draw(self, svg: SVG) -> str:
        """Draw."""
        del svg  # Unused here.
        return ''


class Color(Enum):
    """Predefined colors for SVG elements.

    These should be calibrated to have equal prominence when displayed
    over the BG color in both light and dark modes.
    """

    # Values here correspond to CSS entries (red -> .redstroke, etc).
    MONO = 'mono'  # Black in light mode, White in dark mode.
    RED = 'red'
    GREEN = 'green'
    BLUE = 'blue'
    YELLOW = 'yellow'
    MAGENTA = 'magenta'
    TEAL = 'teal'


class WorldMap(Drawable):
    """Draw SVG based world maps."""

    @dataclass
    class Point:
        """Defines a point on a world-map."""

        latlong: tuple[float, float]
        color: Color = Color.MONO
        opacity: float = 1.0
        radius: float = 3.0

    @dataclass
    class Text:
        """Defines a point on a world-map."""

        latlong: tuple[float, float]
        value: str
        color: Color = Color.MONO
        opacity: float = 1.0
        scale: float = 1.0

    @dataclass
    class Line:
        """Defines a line on a world-map."""

        latlongs: tuple[tuple[float, float], tuple[float, float]]
        color: Color = Color.MONO
        opacity: float = 1.0
        thickness: float = 1.0

    def __init__(
        self,
        elems: Sequence[Point | Text | Line] | None = None,
        elemscale: float = 1.0,
    ):
        self.elems = elems if elems is not None else []
        self.elemscale = elemscale

    @override
    def draw(self, svg: SVG) -> str:
        """Draw."""
        # pylint: disable=too-many-locals
        from efro.util import linearstep

        out = ''
        out += (
            f'<image width="{svg.width}" height="{svg.height}"'
            f' preserveAspectRatio="none"'
            f' href="static/world_map.svg"/>'
        )

        # Filter out exact copies.
        # UPDATE: Not doing this for now; often want to draw multiple
        # times at the same place.
        # de_duped_pts = {p.latlong: p for p in self.points}

        for elem in self.elems:
            if isinstance(elem, self.Point):
                x = svg.width * linearstep(-180.0, 180.0, elem.latlong[1])
                y = svg.height * linearstep(90.0, -90.0, elem.latlong[0])
                opc = (
                    ''
                    if math.isclose(elem.opacity, 1.0)
                    else f' opacity="{elem.opacity:.2f}"'
                )
                radius = elem.radius * self.elemscale
                out += (
                    f'<circle class="{elem.color.value}fill"'
                    f' cx="{x}" cy="{y}" r="{radius}"{opc}/>'
                )
            elif isinstance(elem, self.Line):
                px1 = svg.width * linearstep(-180.0, 180.0, elem.latlongs[0][1])
                py1 = svg.height * linearstep(90.0, -90.0, elem.latlongs[0][0])
                px2 = svg.width * linearstep(-180.0, 180.0, elem.latlongs[1][1])
                py2 = svg.height * linearstep(90.0, -90.0, elem.latlongs[1][0])
                opc = (
                    ''
                    if math.isclose(elem.opacity, 1.0)
                    else f' stroke-opacity="{elem.opacity:.2f}"'
                )
                thickness = elem.thickness * self.elemscale
                out += (
                    f'<line class="{elem.color.value}stroke"'
                    f' x1="{px1}" y1="{py1}" x2="{px2}" y2="{py2}"'
                    f' stroke-width="{thickness}"{opc}/>'
                )

                # DEBUG - show lengths
                if bool(False):
                    cntx = (px1 + px2) * 0.5
                    cnty = (py1 + py2) * 0.5
                    dfx = px2 - px1
                    dfy = py2 - py1
                    llen = math.sqrt(dfx * dfx + dfy * dfy)
                    out += (
                        f'<text class="{elem.color.value}fill"'
                        f' x="{cntx:.1f}" y="{cnty:.1f}" >'
                        f'{llen:.0f}</text>\n'
                    )
            elif isinstance(elem, self.Text):
                x = svg.width * linearstep(-180.0, 180.0, elem.latlong[1])
                y = svg.height * linearstep(90.0, -90.0, elem.latlong[0])
                opc = (
                    ''
                    if math.isclose(elem.opacity, 1.0)
                    else f' opacity="{elem.opacity:.2f}"'
                )
                scl = 100.0 * elem.scale * self.elemscale
                out += (
                    f'<text class="{elem.color.value}fill"'
                    f' x="{x}" y="{y}" font-size="{scl:.2f}%">'
                    f'{elem.value}</text>'
                )

            else:
                raise RuntimeError(f'Invalid elem found: {type(elem)}')
        return out


class Graph(Drawable):
    """Draw simple SVG based graphs.

    Graphs can be nested (to show individual server-node graphs within an
    overall fleet graph for example).
    Note that Y coords are flipped in the graphing system; y==0 is at
    the bottom of the graph while it is at the top of the SVG space.
    """

    @dataclass
    class Curve:
        """Defines a curve on a graph."""

        samples: list[tuple[float, float]] = field(default_factory=list)
        bottom_samples: list[tuple[float, float]] | None = None
        color: Color = Color.MONO
        thickness: float = 1.0
        label: str | None = None
        fill_under: bool = False
        fill_over: bool = False

        # If present, automatic graph y range calculation will include
        # This range. (does not apply if the graph has a specific y-range)
        base_y_range: tuple[float, float] | None = None

        # Overrides standard opacities if present.
        fill_opacity: float | None = None
        line_opacity: float | None = None

    @dataclass
    class Point:
        """Defines a point drawn in a graph."""

        pos: tuple[float, float]
        color: Color = Color.MONO
        radius: float = 2.0

    def __init__(
        self,
        x_range: tuple[float, float] | None = None,
        y_range: tuple[float, float] | None = None,
        elems: Sequence[Graph.Curve | Graph.Point | Graph] | None = None,
        pos: tuple[float, float] | None = None,
        size: tuple[float, float] | None = None,
        tint: Color | None = None,
        grid: bool = True,
        label: str | None = None,
        label_short: str | None = None,
        show_x_values: bool = False,
        show_y_values: bool = False,
        url: str | None = None,
        x_values_precision: int | None = None,
        y_values_precision: int | None = None,
        border: float = 4.0,
        border_top: float = 0.0,
        border_right: float = 0.0,
        border_bottom: float = 0.0,
        border_left: float = 0.0,
        inset_x: float = 0.0,
        inset_y: float = 1.0,
    ) -> None:
        # pylint: disable=too-many-locals
        self.elems = [] if elems is None else list(elems)
        self.x_range = x_range
        self.y_range = y_range
        self.pos = pos
        self.size = size
        self.tint = tint
        self.grid = grid
        self.label = label
        self.label_short = label_short
        self.url = url
        self.show_x_values = show_x_values
        self.show_y_values = show_y_values
        self.x_values_precision = x_values_precision
        self.y_values_precision = y_values_precision
        self.border = border
        self.border_top = border_top
        self.border_right = border_right
        self.border_bottom = border_bottom
        self.border_left = border_left
        self.inset_x = inset_x
        self.inset_y = inset_y
        self.uuid: str | None = None

        # Draw-vars:

        # Graph transform (where graph range maps to in view space)
        self.view_tx = 0.0
        self.view_ty = 0.0
        self.view_sx = 0.0
        self.view_sy = 0.0

        # Bounds transform:
        # (generally a few pixels outset from graph transform so min/max
        # graph values can be clearly seen)
        self.bounds_tx = 0.0
        self.bounds_ty = 0.0
        self.bounds_sx = 0.0
        self.bounds_sy = 0.0

        # Clipping (same as bounds but clipped to parent)
        self.clip_l = 0.0
        self.clip_r = 0.0
        self.clip_t = 0.0
        self.clip_b = 0.0

    def transform_to_view(
        self, point: tuple[float, float]
    ) -> tuple[float, float]:
        """Transform a graph-space point into view space."""
        assert self.x_range is not None
        assert self.y_range is not None

        return (
            self.view_tx
            + self.view_sx
            * (
                (point[0] - self.x_range[0])
                / (self.x_range[1] - self.x_range[0])
            ),
            self.view_ty
            + self.view_sy
            * (
                1.0
                - (point[1] - self.y_range[0])
                / (self.y_range[1] - self.y_range[0])
            ),
        )

    @property
    def clip_id(self) -> str:
        """clipPath id used to clip to this graph when drawing."""
        # Note: originally was using id(self) but it seems its possible
        # for our pointer to be reused which can lead to problems.
        assert self.uuid is not None
        return self.uuid

    def _reset_uuid(self) -> None:
        self.uuid = str(uuid.uuid4())

    def _define_clip(self) -> str:
        clip_width = self.clip_r - self.clip_l
        clip_height = self.clip_b - self.clip_t
        return (
            f'<defs>\n'
            f'  <clipPath id="{self.clip_id}">\n'
            f'    <rect x="{self.clip_l}" y="{self.clip_t}"'
            f' width="{clip_width}" height="{clip_height}" />\n'
            f'  </clipPath>\n'
            f'</defs>\n'
        )

    def use_clip(self) -> str:
        """Return attr def to apply our clip."""
        return f' clip-path="url(#{self.clip_id})"'

    @override
    def draw(self, svg: SVG) -> str:
        """Draw the graph."""

        # We are the top level graph; calc our pos/size in view space.
        self.bounds_tx = self.border + self.border_left
        self.bounds_ty = self.border + self.border_top
        self.bounds_sx = svg.width - (
            2.0 * self.border
            + (26.0 if self.show_y_values else 0.0)
            + self.border_left
            + self.border_right
        )
        self.bounds_sy = svg.height - (
            2.0 * self.border
            + (17.0 if self.show_x_values else 0.0)
            + self.border_top
            + self.border_bottom
        )

        self.view_tx = self.bounds_tx + self.inset_x
        self.view_ty = self.bounds_ty + self.inset_y
        self.view_sx = self.bounds_sx - 2.0 * self.inset_x
        self.view_sy = self.bounds_sy - 2.0 * self.inset_y

        self.clip_l = self.bounds_tx
        self.clip_r = self.bounds_tx + self.bounds_sx
        self.clip_t = self.bounds_ty
        self.clip_b = self.bounds_ty + self.bounds_sy

        out = ''
        out += self.draw_with_parent(None)
        return out

    def calc_x_range(self) -> tuple[float, float]:
        """Calculate min/max values of our contained children."""
        xmins: list[float] = []
        xmaxes: list[float] = []
        for elem in self.elems:
            if isinstance(elem, self.Curve):
                if elem.samples:
                    xmins.append(elem.samples[0][0])
                    xmaxes.append(elem.samples[-1][0])
                if elem.bottom_samples:
                    xmins.append(elem.bottom_samples[0][0])
                    xmaxes.append(elem.bottom_samples[-1][0])
            elif isinstance(elem, self.Point):
                xmins.append(elem.pos[0])
                xmaxes.append(elem.pos[0])
            elif isinstance(elem, Graph):
                if elem.pos is not None and elem.size is not None:
                    xmins.append(elem.pos[0])
                    xmaxes.append(elem.pos[0] + elem.size[0])
        xmin = min(xmins) if xmins else 0.0
        xmax = max(xmaxes) if xmaxes else xmin
        assert xmax >= xmin
        return xmin, xmax

    def calc_y_range(self) -> tuple[float, float]:
        """Calculate min/max values of our contained children."""
        ymins: list[float] = []
        ymaxes: list[float] = []
        for elem in self.elems:
            if isinstance(elem, self.Curve):
                if elem.base_y_range is not None:
                    ymins.append(elem.base_y_range[0])
                    ymaxes.append(elem.base_y_range[1])
                for sample in elem.samples:
                    ymins.append(sample[1])
                    ymaxes.append(sample[1])
                if elem.bottom_samples:
                    for sample in elem.bottom_samples:
                        ymins.append(sample[1])
                        ymaxes.append(sample[1])

            elif isinstance(elem, self.Point):
                ymins.append(elem.pos[1])
                ymaxes.append(elem.pos[1])

            elif isinstance(elem, Graph):
                if elem.pos is not None and elem.size is not None:
                    ymins.append(elem.pos[1])
                    ymaxes.append(elem.pos[1] + elem.size[1])

        ymin = min(ymins) if ymins else 0.0
        ymax = max(ymaxes) if ymaxes else ymin
        assert ymax >= ymin
        return ymin, ymax

    def draw_with_parent(self, parent: Graph | None) -> str:
        """Common drawing for both top level and child graphs."""
        out = ''

        # Calc our display range if it wasn't provided.
        if self.x_range is None:
            self.x_range = self.calc_x_range()
        if self.y_range is None:
            self.y_range = self.calc_y_range()

            # Add a bit of buffer above/below min/max values.
            # yoffs = (self.y_range[1] - self.y_range[0]) * 0.05
            # self.y_range = (self.y_range[0] - yoffs, self.y_range[1] + yoffs)

        # Make sure we're showing non-zero ranges.
        self.x_range = (
            self.x_range[0],
            max(self.x_range[0] + 0.0000001, self.x_range[1]),
        )
        self.y_range = (
            self.y_range[0],
            max(self.y_range[0] + 0.0000001, self.y_range[1]),
        )

        if self.url:
            out += f'<a href="{self.url}">\n'

        # Generate our UUID each time we draw; prevents problems if
        # we duplicate graphs for drawing.
        self._reset_uuid()
        out += self._define_clip()
        out += self._draw_grid(parent)
        out += self._draw_tint_rect()

        # Draw our elems.
        for elem in self.elems:
            if isinstance(elem, self.Curve):
                out += self._draw_curve(elem, parent)
            elif isinstance(elem, self.Point):
                out += self._draw_point(elem)
                x, y = self.transform_to_view(elem.pos)
                out += (
                    f'<circle class="{elem.color.value}fill"'
                    f' cx="{x:.1f}" cy="{y:.1f}"'
                    f' r="{elem.radius:.1f}" stroke="none"'
                    f' {self.use_clip()} />'
                )
            elif isinstance(elem, Graph):
                out += self._draw_subgraph(elem)
            else:
                raise RuntimeError(f'Got unexpected elem type {type(elem)}')

        out += self._draw_x_values(parent)
        out += self._draw_y_values(parent)
        out += self._draw_label(parent)
        out += self._draw_curve_labels(parent)
        out += self._draw_graph_bounds(parent)
        if self.url:
            out += '</a>\n'
        return out

    def _draw_label(self, parent: Graph | None) -> str:
        if not self.label:
            if self.label_short:
                raise RuntimeError("Can't provide label_short but not label.")
            return ''
        label_short = (
            self.label_short if self.label_short is not None else self.label
        )
        out = ''
        buffer = 4.0 if parent is None else 2.0

        # It seems that (on safari at least) clipping sometimes fails to work
        # for very thin sub-graphs. Let's completely avoid drawing in those
        # cases since we'd be mostly unreadable anyway.
        if self.bounds_sx < 2.0:
            return ''

        # Draw bigger for top-level graphs, smaller for embedded ones,
        # and really small for embedded ones with limited space.
        scl, label = (
            ('', self.label)
            if parent is None
            else (
                ('font-size="75%"', self.label)
                if (self.bounds_sx >= 50.0 and self.bounds_sy >= 15)
                else ('font-size="50%"', label_short)
            )
        )

        # If we're starting off the left of the screen, nudge right.
        # Note: this will be ugly if we ever have lots of stuff completely
        # off the left side; may need to do something smarter then.
        x = max(self.clip_l, self.bounds_tx) + buffer
        y = self.bounds_ty + buffer
        label_esc = html.escape(label)
        out += (
            f'<text class="regulartxtfill" x="{x:.1f}" y="{y:.1f}"'
            f'{scl}{self.use_clip()}'
            f' alignment-baseline="hanging">'
            f'{label_esc}</text>\n'
        )
        return out

    def _draw_curve_labels(self, parent: Graph | None) -> str:
        # Currently only draw for top level graphs.
        if parent:
            return ''
        out = ''

        labels: dict[str, Color] = {}
        # Gather labels from all our curves.
        for elem in self.elems:
            if isinstance(elem, self.Curve) and elem.label is not None:
                labels[elem.label] = elem.color

        boxsz = 8
        x = self.bounds_tx + 8
        y = self.bounds_ty + (22 if self.label else 8)
        for name, color in labels.items():
            fillcls = f'{color.value}fill'
            out += (
                f'<rect class="{fillcls}" x="{x:.1f}"'
                f' y="{y:.1f}" width="{boxsz:.1f}"'
                f' height="{boxsz:.1f}" stroke="none"/>\n'
            )
            scl = 'font-size="75%"'
            txt_x = x + boxsz + 4.0
            txt_y = y + boxsz - 1.0
            out += (
                f'<text class="{fillcls}"'
                f' x="{txt_x:.1f}" y="{txt_y:.1f}" {scl} >'
                f'{name}</text>\n'
            )

            y += 14

        return out

    def _automatic_precision(self, valuerange: float, steps: int) -> int:
        """Calc a nice precision for when displaying a range."""
        subdivchange = valuerange / steps
        if bool(False):
            print(f'FOR STEPS {steps} CHANGERATE IS {subdivchange}')
        return 0 if subdivchange > 5 else 1 if subdivchange > 1 else 2

    def _draw_x_values(self, parent: Graph | None) -> str:
        # pylint: disable=too-many-locals
        assert self.x_range is not None
        if not self.show_x_values:
            return ''
        out = ''
        buffer = 3.0
        scl = 'font-size="50%"' if parent is not None else 'font-size="75%"'
        fillcls = (
            'graphoutlinefill' if parent is None else 'subgraphoutlinefill'
        )

        samples: list[tuple[float, float, str, bool]] = [
            (self.view_tx, self.x_range[0], 'start', True)
        ]
        subdivs = int(self.view_sx / 100)
        for i in range(subdivs):
            amt = (i + 1) / (subdivs + 1)
            samples.append(
                (
                    self.view_tx + self.view_sx * amt,
                    self.x_range[1] * amt + (1.0 - amt) * self.x_range[0],
                    'middle',
                    True,
                )
            )
        samples.append(
            (self.view_tx + self.view_sx, self.x_range[1], 'middle', True)
        )

        if self.x_values_precision is not None:
            xvp = self.x_values_precision
        else:
            xvp = self._automatic_precision(
                self.x_range[1] - self.x_range[0], subdivs
            )
        for x, val, anchor, do_box in samples:
            box_width = 1.0
            box_height = 5.0
            if do_box:
                box_x = x - box_width * 0.5
                box_y = self.bounds_ty + self.bounds_sy
                out += (
                    f'<rect class="{fillcls}" x="{box_x:.1f}"'
                    f' y="{box_y:.1f}" width="{box_width:.1f}"'
                    f' height="{box_height:.1f}" stroke="none"/>\n'
                )
            text_x = x
            text_y = self.bounds_ty + self.bounds_sy + box_height + buffer
            label_esc = f'{val:.{xvp}f}'
            out += (
                f'<text class="{fillcls}"'
                f' x="{text_x:.1f}" y="{text_y:.1f}" {scl}'
                f' text-anchor="{anchor}"'
                f' alignment-baseline="hanging">'
                f'{label_esc}</text>\n'
            )
        return out

    def _draw_y_values(self, parent: Graph | None) -> str:
        # pylint: disable=too-many-locals
        assert self.y_range is not None
        if not self.show_y_values:
            return ''
        out = ''
        buffer = 3.0
        scl = 'font-size="50%"' if parent is not None else 'font-size="75%"'
        fillcls = (
            'graphoutlinefill' if parent is None else 'subgraphoutlinefill'
        )

        samples: list[tuple[float, float, str, bool]] = [
            (
                self.view_ty,
                self.y_range[1],
                (
                    'hanging'
                    if self.border + self.border_top < 4.0
                    else 'middle'
                ),
                True,
            )
        ]
        subdivs = int(self.view_sy / 33)
        for i in range(subdivs):
            amt = (i + 1) / (subdivs + 1)
            samples.append(
                (
                    self.view_ty + self.view_sy * amt,
                    self.y_range[0] * amt + (1.0 - amt) * self.y_range[1],
                    'middle',
                    True,
                )
            )
        samples.append(
            (self.view_ty + self.view_sy, self.y_range[0], 'middle', True)
        )

        if self.y_values_precision is not None:
            yvp = self.y_values_precision
        else:
            yvp = self._automatic_precision(
                self.y_range[1] - self.y_range[0], subdivs
            )
        for y, val, align, do_box in samples:
            box_width = 5.0
            box_height = 1.0
            if do_box:
                box_x = self.bounds_tx + self.bounds_sx
                box_y = y - box_height * 0.5
                out += (
                    f'<rect class="{fillcls}" x="{box_x:.1f}"'
                    f' y="{box_y:.1f}" width="{box_width:.1f}"'
                    f' height="{box_height:.1f}" stroke="none"/>\n'
                )
            text_x = self.bounds_tx + self.bounds_sx + box_width + buffer
            text_y = y
            label_esc = f'{val:.{yvp}f}'
            out += (
                f'<text class="{fillcls}"'
                f' x="{text_x:.1f}" y="{text_y:.1f}" {scl}'
                f' alignment-baseline="{align}"'
                f'>{label_esc}</text>\n'
            )
        return out

    def _draw_grid(self, parent: Graph | None) -> str:
        out = ''
        # Draw our background grid (only for top level graphs).
        if self.grid and parent is None:
            grid_uuid = str(uuid.uuid4())
            out += (
                f'<defs>\n'
                f'  <pattern id="smallGrid{grid_uuid}" width="10" height="10"'
                f'     patternUnits="userSpaceOnUse">\n'
                f'  <path class="gridsmall" d="M 10 0 L 0 0 0 10" fill="none"'
                f'    stroke-width="2"/>\n'
                f'  </pattern>\n'
                # f'  <pattern id="grid" width="50" height="50"'
                # f'    patternUnits="userSpaceOnUse">\n'
                # f'  <rect width="50" height="50" fill="url(#smallGrid)"/>\n'
                # f' <path class="gridlarge" d="M 50 0 L 0 0 0 50" fill="none"'
                # f'    stroke-width="2"/>\n'
                # f'  </pattern>\n'
                f'</defs>\n'
                f'<rect x="{self.bounds_tx}"'
                f' y="{self.bounds_ty}" width="{self.bounds_sx}"'
                f' height="{self.bounds_sy}"'
                f' fill="url(#smallGrid{grid_uuid})" stroke="none"/>\n'
            )

        return out

    def _draw_tint_rect(self) -> str:
        out = ''
        if self.tint is not None:
            out += (
                f'<rect class="{self.tint.value}graphtint"'
                f' x="{self.bounds_tx}"'
                f' y="{self.bounds_ty}" width="{self.bounds_sx}"'
                f' height="{self.bounds_sy}"{self.use_clip()}'
                f' stroke="none"/>\n'
            )
        return out

    def _draw_graph_bounds(self, parent: Graph | None) -> str:
        # Draw the graph bounds.
        # Clip to our parent bounds but not our own; we want our outline to be
        # able to creep outside of our own bounds.
        outlinecls = 'graphoutline' if parent is None else 'subgraphoutline'
        stroke_width = 1.0
        parentclip = '' if parent is None else parent.use_clip()

        # We used to draw fully outside of our bounds to avoid obscuring values
        # but we now have inset for that purpose. So let's just draw exactly
        # on our bounds.
        if parent is None and bool(False):
            # Draw outlines outside our graph area when have no parent
            # (avoids obscuring our values)
            posx = self.bounds_tx - 0.5 * stroke_width
            posy = self.bounds_ty - 0.5 * stroke_width
            sizex = self.bounds_sx + 1.0 * stroke_width
            sizey = self.bounds_sy + 1.0 * stroke_width
        else:
            # Draw outlines exactly on our bounds, allowing lines to bleed
            # outside and inside.
            posx = self.bounds_tx
            posy = self.bounds_ty
            sizex = self.bounds_sx
            sizey = self.bounds_sy
        return (
            f'<rect class="{outlinecls}" x="{posx}"'
            f' y="{posy}" width="{sizex}"'
            f' height="{sizey}" stroke-width="{stroke_width}"'
            f' {parentclip} />\n'
        )

    def _draw_curve(self, curve: Curve, parent: Graph | None) -> str:
        # pylint: disable=too-many-branches
        assert self.y_range is not None
        out = ''

        # Gracefully handle empty curves.
        if len(curve.samples) < 1:
            return out

        if curve.bottom_samples is not None:
            if len(curve.bottom_samples) != len(curve.samples):
                raise RuntimeError('Bottom samples and samples len mismatch')
            if curve.fill_under or curve.fill_over:
                raise RuntimeError(
                    "Can't pass fill_under or fill_over with bottom_samples"
                )
        # Optional area shading.
        if (
            curve.fill_under
            or curve.fill_over
            or curve.bottom_samples is not None
        ):
            fill_opacity = (
                curve.fill_opacity
                if curve.fill_opacity is not None
                else 0.1 if parent is None else 0.07
            )
            out += (
                f'<polyline class="{curve.color.value}fill"'
                f' stroke="none" opacity="{fill_opacity:.2f}"'
                f'{self.use_clip()} points="'
            )
            if curve.bottom_samples is not None:
                samples = curve.samples + list(reversed(curve.bottom_samples))
            else:
                # Due to inset we see an edge if we don't extend past y_range;
                # Calc how far in graph space we need to extend.
                extend = (
                    1.0
                    * self.inset_y
                    * ((self.y_range[1] - self.y_range[0]) / self.view_sy)
                )
                y_edge = (
                    self.y_range[0] - extend
                    if curve.fill_under
                    else self.y_range[1] + extend
                )

                # Add samples at the edge of the graph below the first and
                # last points on the curve.
                samples = (
                    [(curve.samples[0][0], y_edge)]
                    + curve.samples
                    + [(curve.samples[-1][0], y_edge)]
                )

            # Run some super-simple filtering on the samples to reduce
            # redundant datapoints in horizontal lines.
            if bool(True):
                samples_filtered: list[tuple[float, float]] = []
                sval: float | None = None
                for i in range(len(samples) - 1):
                    # Copy the sample only if its y value has changed
                    # significantly since the last sample we wrote (sval) or
                    # if the y value of the next sample will be changing.
                    if (
                        sval is None
                        or not math.isclose(sval, samples[i][1])
                        or not math.isclose(sval, samples[i + 1][1])
                    ):
                        sval = samples[i][1]
                        samples_filtered.append(samples[i])
                # Always copy the last sample.
                samples_filtered.append(samples[-1])
            else:
                samples_filtered = samples

            for i, sample in enumerate(samples_filtered):
                x, y = self.transform_to_view(sample)
                if i != 0:
                    out += ' '
                out += f'{x:.1f},{y:.1f}'
            out += '"/>\n'
        line_opacity = (
            curve.line_opacity
            if curve.line_opacity is not None
            else 1.0 if parent is None else 0.4
        )
        out += (
            f'<polyline class="{curve.color.value}stroke"'
            f' stroke-width="{curve.thickness}" opacity="{line_opacity:.2f}"'
            f' fill="none"{self.use_clip()} points="'
        )
        samples = (
            (curve.samples + list(reversed(curve.bottom_samples)))
            if curve.bottom_samples is not None
            else curve.samples
        )
        for i, sample in enumerate(samples):
            x, y = self.transform_to_view(sample)
            if i != 0:
                out += ' '
            out += f'{x:.1f},{y:.1f}'
        out += '"/>\n'
        return out

    def _draw_point(self, point: Point) -> str:
        out = ''
        x, y = self.transform_to_view(point.pos)
        out += (
            f'<circle class="{point.color.value}fill"'
            f' cx="{x:.1f}" cy="{y:.1f}"'
            f' r="{point.radius:.1f}" stroke="none"{self.use_clip()} />'
        )
        return out

    def _draw_subgraph(self, subgraph: Graph) -> str:
        out = ''
        # Set up the child's view coords based on ours.
        assert subgraph.pos is not None and subgraph.size is not None
        subgraph.bounds_tx, subgraph.bounds_ty = self.transform_to_view(
            (subgraph.pos[0], subgraph.pos[1] + subgraph.size[1])
        )
        x, y = self.transform_to_view(
            (subgraph.pos[0] + subgraph.size[0], subgraph.pos[1])
        )
        subgraph.bounds_sx = x - subgraph.bounds_tx
        subgraph.bounds_sy = y - subgraph.bounds_ty

        subgraph.view_tx = subgraph.bounds_tx + subgraph.inset_x
        subgraph.view_ty = subgraph.bounds_ty + subgraph.inset_y
        subgraph.view_sx = subgraph.bounds_sx - 2.0 * subgraph.inset_x
        subgraph.view_sy = subgraph.bounds_sy - 2.0 * subgraph.inset_y

        subgraph.clip_l = max(self.clip_l, subgraph.bounds_tx)
        subgraph.clip_r = min(
            self.clip_r, subgraph.bounds_tx + subgraph.bounds_sx
        )
        subgraph.clip_t = max(self.clip_t, subgraph.bounds_ty)
        subgraph.clip_b = min(
            self.clip_b, subgraph.bounds_ty + subgraph.bounds_sy
        )
        out += subgraph.draw_with_parent(self)
        return out
