# Synced from ballistica-master-server.
# EFRO_SYNC_HASH=165174831160324643760974316257674825315
#
"""Communication from bamaster to basn."""

from __future__ import annotations

import datetime
from enum import Enum
from typing import TYPE_CHECKING, Annotated
from dataclasses import dataclass, field

from typing_extensions import override
from efro.log import LogLevel, LogArchive
from efro.message import Message, Response, BoolResponse, MessageProtocol
from efro.dataclassio import ioprepped, IOAttrs
from baserver.location import Location
from baserver.hostedgame import HostedGameDef
from baserver.enums import TaskTestType, ServerNodeTaskState, ServerNodeRunMode
from baserver.cloudbuild import RecipeTypeID, RecipeComponent

if TYPE_CHECKING:
    pass


def get_protocol() -> MessageProtocol:
    """Get the protocol used in this communication.

    Remember to run 'make meta' after any changes here to update the
    auto-generated sender/receiver classes.
    """

    return MessageProtocol(
        message_types={
            3: LogQueryMessage,
            8: ShutdownMessage,
            10: TestErrorMessage,
            11: UpdateAndRestartMessage,
            12: TaskTestMessage,
            13: TestCommMessage,
            14: InvalidateTokenMessage,
            15: TestAIOErrorMessage,
            17: ClearErrorsMessage,
            18: TestScratchMessage,
            19: PingMessage,
            20: TasksQueryMessage,
            22: TaskQueryMessage,
            24: SetRunModeMessage,
            25: WaitStateMessage,
            26: TestLogMessage,
            27: TestPrintMessage,
            30: HostGameMessage,
            31: TaskActionMessage,
            35: TaskRunningMessage,
            36: TaskKillMessage,
            37: ServerNodeSettingsMessage,
            38: HostPingTargetMessage,
            39: HostTransportAgentMessage,
            40: ForwardedToClientMessage,
            41: GetDebugInfoMessage,
            42: CloudConsoleUpdateMessage,
            43: ExecCloudBuildMessage,
        },
        response_types={
            4: LogQueryResponse,
            9: ShutdownResponse,
            21: TasksQueryResponse,
            23: TaskQueryResponse,
            32: TaskActionResponse,
            33: HostGameResponse,
            34: BoolResponse,
            39: HostPingTargetResponse,
            40: HostTransportAgentResponse,
            41: ForwardedToClientResponse,
            42: GetDebugInfoResponse,
            43: CloudConsoleUpdateResponse,
            44: ExecCloudBuildResponse,
        },
        remote_errors_include_stack_traces=True,
    )


@ioprepped
@dataclass
class LogQueryMessage(Message):
    """Ask basn to send us logs."""

    # Task log we want; None implies basn's log.
    task_id: Annotated[int | None, IOAttrs('task')] = None
    max_size: Annotated[int | None, IOAttrs('maxsize')] = None
    min_log_level: Annotated[LogLevel, IOAttrs('minloglevel')] = LogLevel.DEBUG
    contains: Annotated[str | None, IOAttrs('contains')] = None

    @override
    @classmethod
    def get_response_types(cls) -> list[type[Response] | None]:
        return [LogQueryResponse]


@ioprepped
@dataclass
class CommKeyAckMessage(Message):
    """Response to asking basn for logs."""


@ioprepped
@dataclass
class LogQueryResponse(Response):
    """Response to asking basn for logs."""

    log: Annotated[str, IOAttrs('l')] = ''
    truncated_lines: Annotated[int, IOAttrs('t')] = 0
    error: Annotated[str | None, IOAttrs('e')] = None


@ioprepped
@dataclass
class ShutdownMessage(Message):
    """Tell basn to begin the shutdown process."""

    request_complete: Annotated[bool, IOAttrs('c')] = False

    @override
    @classmethod
    def get_response_types(cls) -> list[type[Response] | None]:
        return [ShutdownResponse]


@ioprepped
@dataclass
class ShutdownResponse(Response):
    """Response from basn containing shutdown state."""

    can_complete: Annotated[bool, IOAttrs('cc')]
    is_complete: Annotated[bool, IOAttrs('c')]


@ioprepped
@dataclass
class TestCommMessage(Message):
    """Message for communication testing."""


@ioprepped
@dataclass
class TestScratchMessage(Message):
    """Message for scratch tests."""


@ioprepped
@dataclass
class TestErrorMessage(Message):
    """Message for error tests."""


@ioprepped
@dataclass
class TestAIOErrorMessage(Message):
    """Message for error tests."""


@ioprepped
@dataclass
class ServerNodeSettingsMessage(Message):
    """Informs basn of settings changes/values."""

    hard_capacity: Annotated[float, IOAttrs('hc')]
    cpu_per_load_unit: Annotated[float, IOAttrs('lcpu')]
    mem_per_load_unit: Annotated[float, IOAttrs('lmem')]

    mem_danger_level: Annotated[float, IOAttrs('mdl')]


@ioprepped
@dataclass
class TestLogMessage(Message):
    """Test logging basn-side."""

    level: Annotated[LogLevel, IOAttrs('l')]


@ioprepped
@dataclass
class TestPrintMessage(Message):
    """Test printing basn-side."""


@ioprepped
@dataclass
class UpdateAndRestartMessage(Message):
    """Tell basn to update itself and restart."""

    full: Annotated[bool, IOAttrs('f')]


@ioprepped
@dataclass
class TaskTestMessage(Message):
    """Test tasks."""

    test_type: Annotated[TaskTestType, IOAttrs('tt')]


@ioprepped
@dataclass
class HostGameMessage(Message):
    """Request to host a game."""

    gamedef: Annotated[HostedGameDef, IOAttrs('gamedef')] = field(
        default_factory=HostedGameDef
    )

    @override
    @classmethod
    def get_response_types(cls) -> list[type[Response] | None]:
        return [HostGameResponse]


@ioprepped
@dataclass
class HostGameResponse(Response):
    """Response to a host-game request."""

    task_id: Annotated[int | None, IOAttrs('i')]
    udp_port: Annotated[int | None, IOAttrs('p')]


@ioprepped
@dataclass
class ExecCloudBuildMessage(Message):
    """Request to exec a cloud build."""

    recipe_type_id: Annotated[RecipeTypeID, IOAttrs('recipe')]
    input_data: Annotated[RecipeComponent, IOAttrs('indata')]
    master_server_build_id: Annotated[str, IOAttrs('cbid')]
    master_server_build_uuid: Annotated[str, IOAttrs('cbuuid')]
    load: Annotated[float, IOAttrs('load')]

    @override
    @classmethod
    def get_response_types(cls) -> list[type[Response] | None]:
        return [ExecCloudBuildResponse]


@ioprepped
@dataclass
class ExecCloudBuildResponse(Response):
    """Response to a cloud-build request."""

    task_id: Annotated[int | None, IOAttrs('i')]


@ioprepped
@dataclass
class HostPingTargetMessage(Message):
    """Ask basn to host a ping-target task."""

    duration_seconds: Annotated[float, IOAttrs('ds')]
    load: Annotated[float, IOAttrs('l')] = 0.0

    @override
    @classmethod
    def get_response_types(cls) -> list[type[Response] | None]:
        return [HostPingTargetResponse]


@ioprepped
@dataclass
class HostPingTargetResponse(Response):
    """Response to a host-ping-target request."""

    task_id: Annotated[int | None, IOAttrs('i')]
    port: Annotated[int | None, IOAttrs('p')]


@ioprepped
@dataclass
class ClearErrorsMessage(Message):
    """Tell basn to reset its max-error-level."""


@ioprepped
@dataclass
class PingMessage(Message):
    """Simple ping message."""


@ioprepped
@dataclass
class TasksQueryMessage(Message):
    """Ask basn for info on all its tasks."""

    min_log_level: Annotated[LogLevel, IOAttrs('mll')] = LogLevel.DEBUG

    @override
    @classmethod
    def get_response_types(cls) -> list[type[Response] | None]:
        return [TasksQueryResponse]


@ioprepped
@dataclass
class TasksInfoMessageEntry:
    """Basic info about a single task."""

    task_id: Annotated[int, IOAttrs('i')]
    label: Annotated[str | None, IOAttrs('lb')]
    user_id: Annotated[str | None, IOAttrs('u')]
    launch_time: Annotated[datetime.datetime, IOAttrs('l')]
    duration: Annotated[float | None, IOAttrs('d')]
    state: Annotated[ServerNodeTaskState, IOAttrs('s')]
    load: Annotated[float, IOAttrs('ld')]
    udp_port: Annotated[int | None, IOAttrs('p')]
    error_flag: Annotated[bool, IOAttrs('e')]
    max_log_level: Annotated[LogLevel, IOAttrs('ll')]
    actions: Annotated[list[str], IOAttrs('a')]


@ioprepped
@dataclass
class TasksQueryResponse(Response):
    """Response to a tasks-query."""

    entries: Annotated[list[TasksInfoMessageEntry], IOAttrs('e')] = field(
        default_factory=list
    )


@ioprepped
@dataclass
class TaskKillMessage(Message):
    """Kill a task if it exists."""

    task_id: Annotated[int, IOAttrs('i')]


@ioprepped
@dataclass
class TaskRunningMessage(Message):
    """Queries whether a task exists and is running."""

    task_id: Annotated[int, IOAttrs('i')]

    @override
    @classmethod
    def get_response_types(cls) -> list[type[Response] | None]:
        return [BoolResponse]


@ioprepped
@dataclass
class TaskQueryMessage(Message):
    """Query extended info for a single task."""

    task_id: Annotated[int, IOAttrs('i')]

    @override
    @classmethod
    def get_response_types(cls) -> list[type[Response] | None]:
        return [TaskQueryResponse, None]


@ioprepped
@dataclass
class TaskQueryResponse(Response):
    """Extended info for a single task."""

    task_id: Annotated[int | None, IOAttrs('i')]
    user_id: Annotated[str | None, IOAttrs('u')]
    label: Annotated[str | None, IOAttrs('l')]
    root: Annotated[str | None, IOAttrs('rt')]
    command: Annotated[str | None, IOAttrs('c')]
    state: Annotated[ServerNodeTaskState | None, IOAttrs('st')]
    duration: Annotated[float | None, IOAttrs('d')]
    exit_success: Annotated[bool | None, IOAttrs('es')]
    stdout_stderr: Annotated[str | None, IOAttrs('o')]
    udp_port: Annotated[int | None, IOAttrs('p')]
    notes: Annotated[str | None, IOAttrs('n')]


@ioprepped
@dataclass
class SetRunModeMessage(Message):
    """Tell basn to set the run-mode in its config."""

    run_mode: Annotated[ServerNodeRunMode, IOAttrs('r')]


@ioprepped
@dataclass
class WaitStateMessage(Message):
    """Tell basn to block for a moment until the state has synced to us."""


@ioprepped
@dataclass
class TaskActionMessage(Message):
    """Tell a task on a servernode to run a user action."""

    task_id: Annotated[int, IOAttrs('tid')]
    action: Annotated[str, IOAttrs('a')]

    @override
    @classmethod
    def get_response_types(cls) -> list[type[Response] | None]:
        return [TaskActionResponse]


@ioprepped
@dataclass
class TaskActionResponse(Response):
    """Response from servernode asked to run a task action."""

    output: Annotated[str, IOAttrs('o')]
    center: Annotated[bool, IOAttrs('c')] = True
    auto_return: Annotated[bool, IOAttrs('ar')] = True
    auto_return_wait: Annotated[int, IOAttrs('arw')] = 500

    copy_text: Annotated[str | None, IOAttrs('ct')] = None


@ioprepped
@dataclass
class HostTransportAgentMessage(Message):
    """Ask basn to host a transport-agent task."""

    connection_wait_seconds: Annotated[float, IOAttrs('ds')]
    max_duration_seconds: Annotated[float, IOAttrs('mds')]
    token: Annotated[str, IOAttrs('tk')]
    load: Annotated[float, IOAttrs('l')]

    # Informs basn where the request for a transport-agent originated from.
    location: Annotated[Location | None, IOAttrs('lc')]
    remote_addr: Annotated[str, IOAttrs('r')]
    user_agent_string: Annotated[str, IOAttrs('u')]

    @override
    @classmethod
    def get_response_types(cls) -> list[type[Response] | None]:
        return [HostTransportAgentResponse]


@ioprepped
@dataclass
class HostTransportAgentResponse(Response):
    """Response to a host-transport-agent request."""

    task_id: Annotated[int | None, IOAttrs('i')]
    port: Annotated[int | None, IOAttrs('p')]


@ioprepped
@dataclass
class ForwardedToClientMessage(Message):
    """A message coming from bamaster going to a client."""

    message: Annotated[str, IOAttrs('m')]

    # Local task ID of the V2TransportAgent where client is connected.
    client_task_id: Annotated[int, IOAttrs('c')]

    @override
    @classmethod
    def get_response_types(cls) -> list[type[Response] | None]:
        return [ForwardedToClientResponse]


@ioprepped
@dataclass
class ForwardedToClientResponse(Response):
    """Response to a message forwarded from bamaster to a client."""

    class ErrorType(Enum):
        """Type of error we can report."""

        COMMUNICATION = 'comm'
        UNKNOWN = 'unknown'

    error: Annotated[ErrorType | None, IOAttrs('e')] = None
    response: Annotated[str | None, IOAttrs('r')] = None


@ioprepped
@dataclass
class GetDebugInfoMessage(Message):
    """Ask basn for random debugging info."""

    @override
    @classmethod
    def get_response_types(cls) -> list[type[Response] | None]:
        return [GetDebugInfoResponse]


@ioprepped
@dataclass
class GetDebugInfoResponse(Response):
    """Response from basn containing various debug info."""

    output: Annotated[str, IOAttrs('o')]


@ioprepped
@dataclass
class CloudConsoleUpdateMessage(Message):
    """Messages from cloud-console."""

    # Code to run (if any).
    code: Annotated[str | None, IOAttrs('c')]

    # Next log entry index we're looking for.
    log_index: Annotated[int, IOAttrs('i')]

    # App-uuid associated with the logs we already have.
    # If this doesn't match, log_index is implicitly taken to be 0.
    target_uuid: Annotated[str | None, IOAttrs('u')]

    @override
    @classmethod
    def get_response_types(cls) -> list[type[Response] | None]:
        return [CloudConsoleUpdateResponse]


@ioprepped
@dataclass
class CloudConsoleUpdateResponse(Response):
    """Response to cloud-console."""

    # Instance id of the app.
    target_uuid: Annotated[str, IOAttrs('i')]

    # Current local real-time in the app.
    app_real_time: Annotated[float, IOAttrs('t')]

    # Logs requested.
    log_archive: Annotated[LogArchive, IOAttrs('l')]


@ioprepped
@dataclass
class InvalidateTokenMessage(Message):
    """Tell basn to invalidate a token."""

    token: Annotated[str, IOAttrs('t')]
