# Synced from ballistica-master-server.
# EFRO_SYNC_HASH=23707222296465176900404942300954870941
#
"""Communication from basn to bamaster."""

from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING, Annotated
from dataclasses import dataclass, field

from typing_extensions import override
from efro.log import LogEntry
from efro.dataclassio import ioprepped, IOAttrs
from efro.message import Message, BoolResponse, MessageProtocol, Response
from bacommon.login import LoginType
from baserver.location import Location
from baserver.data import BASNStats, BASNStatsSparse
from baserver.msg import batocloud
from baserver.cloudbuild import RecipeComponent

if TYPE_CHECKING:
    pass


def get_protocol() -> MessageProtocol:
    """Get the protocol used in this communication.

    Remember to run 'make meta' after any changes here to update the
    auto-generated sender/receiver classes.
    """

    return MessageProtocol(
        message_types={
            0: ForwardedClientMessage,
            1: LoginProxyRequestInternalMessage,
            2: LoginProxyStateQueryInternalMessage,
            3: LoginProxyCompleteInternalMessage,
            4: AccountSessionReleaseInternalMessage,
            5: AccountSessionRegisterChannelMessage,
            7: AccountSessionDeleteChannelsMessage,
            8: LogMessage,
            15: ServerNodeStatsMessage,
            16: TestCommResponseMessage,
            17: ServerNodeStatsSparseMessage,
            18: SignInInternalMessage,
            19: CloudFilesExistQueryMessage,
            20: CloudBuildResultsMessage,
        },
        response_types={
            0: ForwardedClientResponse,
            1: LoginProxyRequestInternalResponse,
            2: LoginProxyStateQueryInternalResponse,
            3: AccountSessionRegisterChannelResponse,
            4: SignInInternalResponse,
            5: CloudFilesExistQueryResponse,
            34: BoolResponse,
        },
        remote_errors_include_stack_traces=True,
    )


@ioprepped
@dataclass
class ServerNodeStatsMessage(Message):
    """Informs master of various server node stats values."""

    stats: Annotated[BASNStats, IOAttrs('s')] = field(default_factory=BASNStats)


@ioprepped
@dataclass
class ServerNodeStatsSparseMessage(Message):
    """Informs master of various server node stats values."""

    stats: Annotated[BASNStatsSparse, IOAttrs('s')]


# Note: this is not actually a response; it just is logically.
@ioprepped
@dataclass
class TestCommResponseMessage(Message):
    """Message we ourselves send after receiving a test comm message."""


@ioprepped
@dataclass
class ForwardedClientMessage(Message):
    """A message coming from a client going to bamaster."""

    message: Annotated[str, IOAttrs('m')]

    # Local task ID of the V2TransportAgent that sent this.
    client_task_id: Annotated[int, IOAttrs('c')]

    @override
    @classmethod
    def get_response_types(cls) -> list[type[Response] | None]:
        return [ForwardedClientResponse]


@ioprepped
@dataclass
class ForwardedClientResponse(Response):
    """Response to a message forwarded from client to bamaster."""

    response: Annotated[str, IOAttrs('r')]


@ioprepped
@dataclass
class LoginProxyRequestInternalMessage(Message):
    """Request send to the cloud to ask for a login-proxy."""

    user_agent_string: Annotated[str, IOAttrs('ua', soft_default='')]
    remote_addr: Annotated[str, IOAttrs('a', soft_default='')]
    location: Annotated[Location | None, IOAttrs('l', soft_default=None)]

    @override
    @classmethod
    def get_response_types(cls) -> list[type[Response] | None]:
        return [LoginProxyRequestInternalResponse]


@ioprepped
@dataclass
class LoginProxyRequestInternalResponse(Response):
    """Response to a request for a login proxy."""

    # URL to direct the user to for login.
    url: Annotated[str, IOAttrs('u')]

    # Proxy-Login id for querying results.
    proxyid: Annotated[str, IOAttrs('p')]

    # Proxy-Login key for querying results.
    proxykey: Annotated[str, IOAttrs('k')]


@ioprepped
@dataclass
class LoginProxyStateQueryInternalMessage(Message):
    """Soo.. how is that login proxy going?"""

    proxyid: Annotated[str, IOAttrs('p')]
    proxykey: Annotated[str, IOAttrs('k')]
    buildnum: Annotated[int | None, IOAttrs('b', soft_default=None)]

    @override
    @classmethod
    def get_response_types(cls) -> list[type[Response] | None]:
        return [LoginProxyStateQueryInternalResponse]


@ioprepped
@dataclass
class LoginProxyStateQueryInternalResponse(Response):
    """Here's the info on that login-proxy you asked about, boss."""

    class State(Enum):
        """States a login-proxy can be in."""

        WAITING = 'waiting'
        SUCCESS = 'success'
        FAIL = 'fail'

    state: Annotated[State, IOAttrs('s')]

    # All of this is filled out on success
    token: Annotated[str | None, IOAttrs('tk')]


@ioprepped
@dataclass
class LoginProxyCompleteInternalMessage(Message):
    """Just so you know, we're done with this proxy."""

    proxyid: Annotated[str, IOAttrs('p')]


@ioprepped
@dataclass
class AccountSessionReleaseInternalMessage(Message):
    """We're done using this particular session."""

    token: Annotated[str, IOAttrs('tk')]


@ioprepped
@dataclass
class AccountSessionRegisterChannelMessage(Message):
    """Request a communication channel on behalf of an account session."""

    token: Annotated[str, IOAttrs('t')]

    channel_id: Annotated[int, IOAttrs('h')]
    user_agent_string: Annotated[str, IOAttrs('ua')]
    remote_addr: Annotated[str, IOAttrs('a')]
    location: Annotated[Location | None, IOAttrs('l')]

    task_id: Annotated[int, IOAttrs('i')]

    device_name: Annotated[str, IOAttrs('n')]
    build_number: Annotated[int, IOAttrs('b')]

    app_instance_uuid: Annotated[str, IOAttrs('u')]

    @override
    @classmethod
    def get_response_types(cls) -> list[type[Response] | None]:
        return [AccountSessionRegisterChannelResponse]


@ioprepped
@dataclass
class AccountSessionRegisterChannelResponse(Response):
    """Response as to whether we can use an account session."""

    verified: Annotated[bool, IOAttrs('v')]
    info: Annotated[batocloud.CredentialsCheckResponseInfo | None, IOAttrs('i')]

    channel_id: Annotated[int | None, IOAttrs('c')]


@ioprepped
@dataclass
class AccountSessionDeleteChannelsMessage(Message):
    """Tell the master-server a task is no longer using sessions."""

    task_id: Annotated[int, IOAttrs('i')]

    # session-channel-ids per session.
    channel_ids_per_session: Annotated[dict[str, list[int]], IOAttrs('h')]


@ioprepped
@dataclass
class SignInInternalMessage(Message):
    """Can I sign in please?"""

    login_type: Annotated[LoginType, IOAttrs('l')]
    sign_in_token: Annotated[str, IOAttrs('t')]

    source_task: Annotated[str, IOAttrs('c')]
    description: Annotated[str, IOAttrs('d')]
    apptime: Annotated[float, IOAttrs('at')]

    @override
    @classmethod
    def get_response_types(cls) -> list[type[Response] | None]:
        return [SignInInternalResponse]


@ioprepped
@dataclass
class SignInInternalResponse(Response):
    """Here's that sign-in result you asked for, boss."""

    token: Annotated[str | None, IOAttrs('t')]


@ioprepped
@dataclass
class LogMessage(Message):
    """Sending basn logs to bamaster."""

    log_entries: Annotated[list[LogEntry], IOAttrs('e')]


@ioprepped
@dataclass
class CloudFilesExistQueryMessage(Message):
    """Which of these cloud-files already exist in the cloud?"""

    cloud_file_ids: Annotated[list[str], IOAttrs('f')]

    @override
    @classmethod
    def get_response_types(cls) -> list[type[Response] | None]:
        return [CloudFilesExistQueryResponse]


@ioprepped
@dataclass
class CloudFilesExistQueryResponse(Response):
    """Here's the info on that login-proxy you asked about, boss."""

    cloud_files_exist: Annotated[list[bool], IOAttrs('e')]


@ioprepped
@dataclass
class CloudBuildResultsMessage(Message):
    """Here's how that build went, boss."""

    master_server_build_id: Annotated[str, IOAttrs('b')]
    master_server_build_uuid: Annotated[str, IOAttrs('u')]

    # One of these two should be set.
    output_data: Annotated[RecipeComponent | None, IOAttrs('o')]
    error: Annotated[str | None, IOAttrs('e')]

    @override
    @classmethod
    def get_response_types(cls) -> list[type[Response] | None]:
        return [None]
