# Synced from ballistica-master-server.
# EFRO_SYNC_HASH=72769517188773126193247870514350838200
#
# Copyright (c) 2011-2024 Eric Froemling
"""Communication from client to basn.

IMPORTANT: we embed this on the client side so make sure all names
defined here are unique.
"""

from __future__ import annotations

import datetime
from enum import Enum
from typing import TYPE_CHECKING, Annotated
from dataclasses import dataclass

from typing_extensions import override
from efro.message import MessageProtocol, Message, Response
from efro.dataclassio import will_ioprep, ioprep, IOAttrs
from bacommon.login import LoginType
import bacommon.cloud

if TYPE_CHECKING:
    pass

# Note on @ioprepping:
# We need to pass our globals explicitly since we might be getting
# execed as a string, not a module. For this reason we have to run
# explicit ioprep() calls. And we also insert a @will_ioprep so our
# pylint filters properly expect annotated types to be present.


def get_ba_to_cloud_protocol_full() -> MessageProtocol:
    """Get the protocol used in this communication.

    This is the full protocol containing all message types originating
    from the client regardless of their ultimate destination.

    Remember to run 'make meta' after any changes here to update the
    auto-generated sender/receiver classes.
    """

    return MessageProtocol(
        message_types={
            0: CToBMsg1,
            1: CToBMsg2,
            2: CToBMsg3,
            3: bacommon.cloud.LoginProxyRequestMessage,
            4: bacommon.cloud.LoginProxyStateQueryMessage,
            5: bacommon.cloud.LoginProxyCompleteMessage,
            6: AccountSessionReleaseMessage,
            7: CredentialsCheckMessage,
            8: PostLoginMessage,
            9: bacommon.cloud.WorkspaceFetchMessage,
            10: bacommon.cloud.PingMessage,
            11: AccountChannelCreateMessage,
            12: AccountChannelDestroyMessage,
            13: bacommon.cloud.SignInMessage,
            14: bacommon.cloud.ManageAccountMessage,
            15: bacommon.cloud.MerchAvailabilityMessage,
            16: bacommon.cloud.PromoCodeMessage,
        },
        response_types={
            0: CToBResp1,
            1: CToBResp2,
            2: bacommon.cloud.LoginProxyRequestResponse,
            3: bacommon.cloud.LoginProxyStateQueryResponse,
            5: CredentialsCheckResponse,
            6: PostLoginResponse,
            7: bacommon.cloud.WorkspaceFetchResponse,
            8: bacommon.cloud.PingResponse,
            9: bacommon.cloud.SignInResponse,
            10: bacommon.cloud.ManageAccountResponse,
            11: bacommon.cloud.MerchAvailabilityResponse,
            12: bacommon.cloud.PromoCodeResponse,
        },
        forward_communication_errors=True,
    )


# Message types in our protocol handled by basn.
# Types *NOT* listed here will be passed through to bamaster.
def _get_basn_handled_ba_to_cloud_message_types() -> set[type[Message]]:
    return {
        CToBMsg1,
        CToBMsg2,
        bacommon.cloud.LoginProxyRequestMessage,
        bacommon.cloud.LoginProxyStateQueryMessage,
        bacommon.cloud.LoginProxyCompleteMessage,
        AccountSessionReleaseMessage,
        CredentialsCheckMessage,
        AccountChannelCreateMessage,
        AccountChannelDestroyMessage,
        bacommon.cloud.SignInMessage,
    }


def get_ba_to_cloud_protocol_basn() -> MessageProtocol:
    """Get the basn version of the client-to-basn protocol.

    This includes only the message subset handled by basn.
    """
    base = get_ba_to_cloud_protocol_full()
    basn_handled_types = _get_basn_handled_ba_to_cloud_message_types()
    return MessageProtocol(
        message_types={
            mid: mtp
            for mid, mtp in base.message_types_by_id.items()
            if mtp in basn_handled_types
        },
        response_types={
            rid: rtp
            for rid, rtp in base.response_types_by_id.items()
            if issubclass(rtp, Response)
        },
        forward_communication_errors=True,
    )


def get_ba_to_cloud_protocol_bamaster() -> MessageProtocol:
    """Get the bamaster version of the client-to-basn protocol.

    This includes only the message subset handled by bamaster.
    """
    base = get_ba_to_cloud_protocol_full()
    basn_handled_types = _get_basn_handled_ba_to_cloud_message_types()
    return MessageProtocol(
        message_types={
            mid: mtp
            for mid, mtp in base.message_types_by_id.items()
            if mtp not in basn_handled_types
        },
        response_types={
            rid: rtp
            for rid, rtp in base.response_types_by_id.items()
            if issubclass(rtp, Response)
        },
        forward_communication_errors=True,
    )


# These 2 classes aren't actually part of the protocol; they're just
# standalone types used to establish the communication our protocol uses.
@will_ioprep
@dataclass
class TransportAgentRequest:
    """Request send to bamaster to ask for an agent."""

    # Zones sorted by ping (nearest first).
    zoneids: Annotated[list[str], IOAttrs('z', soft_default_factory=list)]
    build_number: Annotated[int, IOAttrs('b', soft_default=0)]
    current_time: Annotated[
        datetime.datetime | None, IOAttrs('t', soft_default=None)
    ]

    # Old; can kill this when build 20888+ is ubiquitous.
    zoneid: Annotated[str | None, IOAttrs('r')] = None

    # For debugging.
    short_duration: Annotated[bool, IOAttrs('sd', store_default=False)] = False


ioprep(TransportAgentRequest, globalns=globals())


@will_ioprep
@dataclass
class TransportAgentRequestResponse:
    """Response from bamaster for a transport agent request."""

    class Error(Enum):
        """An error type we can return."""

        INTERNAL = 'internal'
        REJECTED = 'rejected'
        CLIENT_TOO_OLD = 'client_too_old'

    # Specifics if anything went wrong.
    error: Annotated[Error | None, IOAttrs('e')]

    address: Annotated[str | None, IOAttrs('a')]
    port: Annotated[int | None, IOAttrs('p')]
    token: Annotated[str | None, IOAttrs('t')]
    cadata: Annotated[str | None, IOAttrs('c')]


ioprep(TransportAgentRequestResponse, globalns=globals())


# Also not protocol classes; sent/received while establishing stream comms.
@will_ioprep
@dataclass
class TransportAgentHandshakeMessage(Message):
    """Info sent to basn when establishing connection."""

    ba_build: Annotated[int, IOAttrs('b')]

    # Added in build 20696; can remove soft-default once that is ubiquitous.
    device_name: Annotated[str, IOAttrs('n', soft_default='?')]
    private_device_uuid: Annotated[str, IOAttrs('u', soft_default='?')]

    # Added in build 20888; can remove soft-default once that is ubiquitous.
    app_instance_uuid: Annotated[str, IOAttrs('au', soft_default='')]


ioprep(TransportAgentHandshakeMessage, globalns=globals())


@will_ioprep
@dataclass
class TransportAgentHandshakeResponse(Response):
    """Info returned from basn in response to handshake."""

    basn_build: Annotated[int, IOAttrs('b')]


ioprep(TransportAgentHandshakeResponse, globalns=globals())


@will_ioprep
@dataclass
class CToBMsg1(Message):
    """Just testing."""

    ival: Annotated[int, IOAttrs('i')]

    @override
    @classmethod
    def get_response_types(cls) -> list[type[Response] | None]:
        return [CToBResp1]


ioprep(CToBMsg1, globalns=globals())


@will_ioprep
@dataclass
class CToBMsg2(Message):
    """Just testing."""

    sval: Annotated[str, IOAttrs('s')]

    @override
    @classmethod
    def get_response_types(cls) -> list[type[Response] | None]:
        return [CToBResp1, CToBResp2]


ioprep(CToBMsg2, globalns=globals())


@will_ioprep
@dataclass
class CToBMsg3(Message):
    """Just testing."""

    sval: Annotated[str, IOAttrs('s')]


ioprep(CToBMsg3, globalns=globals())


@will_ioprep
@dataclass
class CToBResp1(Response):
    """Just testing."""

    bval: Annotated[bool, IOAttrs('b')]


ioprep(CToBResp1, globalns=globals())


@will_ioprep
@dataclass
class CToBResp2(Response):
    """Just testing."""

    fval: Annotated[float, IOAttrs('f')]


ioprep(CToBResp2, globalns=globals())


# Used by builds prior to 20888
@will_ioprep
@dataclass
class CredentialsCheckMessage(Message):
    """Trying to sign in; check my credentials plz."""

    @override
    @classmethod
    def get_response_types(cls) -> list[type[Response] | None]:
        return [CredentialsCheckResponse]


ioprep(CredentialsCheckMessage, globalns=globals())


# NOTE: this data gets forwarded from bamaster through basn to clients
# so we don't have to wait for a basn update to propagate when we add stuff
# (which we would if we were adding it directly to the response class).
@will_ioprep
@dataclass
class CredentialsCheckResponseInfo:
    """Extra info returned with verified credentials."""

    @dataclass
    class LoginInfo:
        """Info about a particular login."""

        login_type: LoginType
        login_id: str
        display_name: str

    tag: Annotated[str, IOAttrs('t')]
    v1login: Annotated[str, IOAttrs('v1')]

    workspaceid: Annotated[str | None, IOAttrs('w')]
    workspacename: Annotated[str | None, IOAttrs('wn')]

    logins: Annotated[list[LoginInfo], IOAttrs('li')]

    accountid: Annotated[str, IOAttrs('a')]


ioprep(CredentialsCheckResponseInfo, globalns=globals())


@will_ioprep
@dataclass
class CredentialsCheckResponse(Response):
    """Checked your credentials; here's your info."""

    verified: Annotated[bool, IOAttrs('v')]

    # DELETE THIS ONCE BA < 20606 IS OFFLINE
    legacy_tag: Annotated[str, IOAttrs('t')]

    # Include all data here; not in the message body itself
    # (so we don't have to wait for basn updates on changes)
    info: Annotated[CredentialsCheckResponseInfo | None, IOAttrs('i')]


ioprep(CredentialsCheckResponse, globalns=globals())


@will_ioprep
@dataclass
class PostLoginMessage(Message):
    """Sent to the cloud just after signing in.

    Can be used to fetch basic account info or whatever.
    """

    @override
    @classmethod
    def get_response_types(cls) -> list[type[Response] | None]:
        return [PostLoginResponse]


ioprep(PostLoginMessage, globalns=globals())


@will_ioprep
@dataclass
class PostLoginResponse(Response):
    """Tell a just-logged-in client what to do."""

    workspaceid: Annotated[
        str | None, IOAttrs('w', store_default=False, soft_default=None)
    ]
    workspacename: Annotated[
        str | None, IOAttrs('wn', store_default=False, soft_default=None)
    ]


ioprep(PostLoginResponse, globalns=globals())


@will_ioprep
@dataclass
class AccountSessionReleaseMessage(Message):
    """We're done using this particular session."""

    token: Annotated[str, IOAttrs('tk')]


ioprep(AccountSessionReleaseMessage, globalns=globals())


@will_ioprep
@dataclass
class AccountChannelCreateMessage(Message):
    """We're done using this particular session."""

    channel_id: Annotated[int, IOAttrs('i')]
    token: Annotated[str, IOAttrs('t')]


ioprep(AccountChannelCreateMessage, globalns=globals())


@will_ioprep
@dataclass
class AccountChannelDestroyMessage(Message):
    """We're done using this particular session."""

    channel_id: Annotated[int, IOAttrs('i')]


ioprep(AccountChannelDestroyMessage, globalns=globals())
