# Synced from ballistica-master-server.
# EFRO_SYNC_HASH=238555670924885183607844458238092995845
#
# Copyright (c) 2011-2024 Eric Froemling
"""Communication from basn to client.

IMPORTANT: we embed this on the client side in the same namespace
as all other embedded code so make sure all names defined here are
unique.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Annotated
from dataclasses import dataclass

from typing_extensions import override
from efro.log import LogArchive
from efro.message import MessageProtocol, Message, Response
from efro.dataclassio import will_ioprep, ioprep, IOAttrs

if TYPE_CHECKING:
    pass

# Note on @ioprepping:
# We need to pass our globals explicitly since we might be getting
# execed as a string, not a module. For this reason we have to run
# explicit ioprep() calls. And we also insert a @will_ioprep so our
# pylint filters properly expect annotated types to be present.


# Not actually message related; just various useful account info we
# send to the user and store locally with them.
@will_ioprep
@dataclass
class AccountCredentials:
    """Various values associated with a logged-in account."""

    token: Annotated[str, IOAttrs('t')]

    # These can go away once build 20927 or newer is ubiquitous.
    legacy_tag: Annotated[str | None, IOAttrs('n')] = None
    legacy_v1login: Annotated[str | None, IOAttrs('l1')] = None


ioprep(AccountCredentials, globalns=globals())


# IMPORTANT: by default, messages are assumed to be coming from bamaster
# and getting forwarded through. To set up a message originating from
# basn, add the message type here.
def _get_basn_handled_b_to_c_message_types() -> set[type[Message]]:
    return {
        BToCMsg1,
        BToCMsg2,
        TransportAgentShutdownWarningMessage,
        TokenInvalidatedMessage,
    }


def get_basn_to_client_protocol_full() -> MessageProtocol:
    """Get the protocol used in this communication.

    This is the full protocol containing all message types received
    by the client regardless of their source. It is used directly
    by the client.

    Remember to run 'make meta' after any changes here to update the
    auto-generated sender/receiver classes.
    """

    return MessageProtocol(
        message_types={
            0: BToCMsg1,
            1: BToCMsg2,
            2: BToCMsg3,
            3: BToCMsg4,
            4: CloudConsoleUpdateMessage,
            5: TransportAgentShutdownWarningMessage,
            6: TokenInvalidatedMessage,
            7: WorkspaceSetupChangedMessage,
        },
        response_types={
            0: BToCResp1,
            1: BToCResp2,
            2: CloudConsoleUpdateResponse,
        },
        # Sender here is trusted to see error details from the client.
        # We could consider taking this further and not even logging
        # the errors on the client, but leaving them on for now in case
        # this would result in errors slipping through the cracks.
        log_remote_errors=True,
        remote_errors_include_stack_traces=True,
        forward_communication_errors=True,
        forward_clean_errors=True,
    )


def get_basn_to_client_protocol_basn() -> MessageProtocol:
    """Get the basn version of the basn-to-client protocol.

    This includes only the message subset sent by basn.
    """
    base = get_basn_to_client_protocol_full()
    basn_handled_types = _get_basn_handled_b_to_c_message_types()
    return MessageProtocol(
        message_types={
            mid: mtp
            for mid, mtp in base.message_types_by_id.items()
            if mtp in basn_handled_types
        },
        response_types={
            rid: rtp
            for rid, rtp in base.response_types_by_id.items()
            if issubclass(rtp, Response)
        },
        # (see notes in full protocol)
        log_remote_errors=True,
        remote_errors_include_stack_traces=True,
        forward_communication_errors=True,
        forward_clean_errors=True,
    )


def get_basn_to_client_protocol_bamaster() -> MessageProtocol:
    """Get the bamaster version of the basn-to-client protocol.

    This includes only the message subset sent by bamaster and forwarded
    on by basn.
    """
    base = get_basn_to_client_protocol_full()
    basn_handled_types = _get_basn_handled_b_to_c_message_types()
    return MessageProtocol(
        message_types={
            mid: mtp
            for mid, mtp in base.message_types_by_id.items()
            if mtp not in basn_handled_types
        },
        response_types={
            rid: rtp
            for rid, rtp in base.response_types_by_id.items()
            if issubclass(rtp, Response)
        },
        # (see notes in full protocol)
        log_remote_errors=True,
        remote_errors_include_stack_traces=True,
        forward_communication_errors=True,
        forward_clean_errors=True,
    )


@will_ioprep
@dataclass
class BToCMsg1(Message):
    """Just testing."""

    ival: Annotated[int, IOAttrs('i')]

    @override
    @classmethod
    def get_response_types(cls) -> list[type[Response] | None]:
        return [BToCResp1]


ioprep(BToCMsg1, globalns=globals())


@will_ioprep
@dataclass
class BToCMsg2(Message):
    """Just testing."""

    sval: Annotated[str, IOAttrs('s')]

    @override
    @classmethod
    def get_response_types(cls) -> list[type[Response] | None]:
        return [BToCResp1, BToCResp2]


ioprep(BToCMsg2, globalns=globals())


@will_ioprep
@dataclass
class BToCMsg3(Message):
    """Just testing."""

    sval: Annotated[str, IOAttrs('s')]


ioprep(BToCMsg3, globalns=globals())


@will_ioprep
@dataclass
class BToCMsg4(Message):
    """Just testing."""

    sval: Annotated[str, IOAttrs('s')]


ioprep(BToCMsg4, globalns=globals())


@will_ioprep
@dataclass
class BToCResp1(Response):
    """Just testing."""

    bval: Annotated[bool, IOAttrs('b')]


ioprep(BToCResp1, globalns=globals())


@will_ioprep
@dataclass
class BToCResp2(Response):
    """Just testing."""

    fval: Annotated[float, IOAttrs('f')]


ioprep(BToCResp2, globalns=globals())


@will_ioprep
@dataclass
class CloudConsoleUpdateMessage(Message):
    """Cloud-console asking us to run something and/or give it output."""

    # Code to run (if any).
    code: Annotated[str | None, IOAttrs('c')]

    # Next log entry index we're looking for.
    log_index: Annotated[int, IOAttrs('i')]

    # uuid associated with the logs we already have.
    # If this doesn't match, log_index is implicitly taken to be 0.
    target_uuid: Annotated[str | None, IOAttrs('u')]

    @override
    @classmethod
    def get_response_types(cls) -> list[type[Response] | None]:
        return [CloudConsoleUpdateResponse]


ioprep(CloudConsoleUpdateMessage, globalns=globals())


@will_ioprep
@dataclass
class CloudConsoleUpdateResponse(Response):
    """Returning logs/etc to cloud console."""

    # Instance id of the app.
    target_uuid: Annotated[str, IOAttrs('i')]

    # Current local real-time in the app.
    app_real_time: Annotated[float, IOAttrs('t')]

    # Logs requested.
    log_archive: Annotated[LogArchive, IOAttrs('l')]


ioprep(CloudConsoleUpdateResponse, globalns=globals())


@will_ioprep
@dataclass
class TransportAgentShutdownWarningMessage(Message):
    """Our transport agent is about to die so we should get a new one."""


ioprep(TransportAgentShutdownWarningMessage, globalns=globals())


@will_ioprep
@dataclass
class TokenInvalidatedMessage(Message):
    """A token is no longer valid (for clients)."""

    token: Annotated[str, IOAttrs('t')]


ioprep(TokenInvalidatedMessage, globalns=globals())


@will_ioprep
@dataclass
class WorkspaceSetupChangedMessage(Message):
    """Client's active workspace situation has changed."""


ioprep(WorkspaceSetupChangedMessage, globalns=globals())
