# Synced from ballistica-master-server.
# EFRO_SYNC_HASH=116995653874504791891715988362622925757
#
"""Communication from bamaster to basn."""

from __future__ import annotations

from typing import TYPE_CHECKING, Annotated
from dataclasses import dataclass

from typing_extensions import override
from baserver.enums import V1Store
from efro.message import Message, MessageProtocol, StringResponse, BoolResponse
from efro.dataclassio import ioprepped, IOAttrs

if TYPE_CHECKING:
    from efro.message import Response


def get_protocol() -> MessageProtocol:
    """Get the protocol used in this communication.

    Remember to run 'make meta' after any changes here to update the
    auto-generated sender/receiver classes.
    """

    return MessageProtocol(
        message_types={
            27: GetLoginCodeMessage,
            29: AccountSyncMessage,
            40: PrivatePartyQueryMessage,
            41: PrivatePartyConnectMessage,
            42: PrivatePartyStartMessage,
            43: PrivatePartyStopMessage,
            44: AdVerificationMessage,
            47: V1IAPVerifyMessage,
            48: SendMailMessage,
        },
        response_types={
            28: StringResponse,
            34: BoolResponse,
        },
        remote_errors_include_stack_traces=True,
    )


@ioprepped
@dataclass
class GetLoginCodeMessage(Message):
    """Fetch a login code for an account."""

    account_id: Annotated[str, IOAttrs('a')]

    @override
    @classmethod
    def get_response_types(cls) -> list[type[Response] | None]:
        return [StringResponse]


@ioprepped
@dataclass
class AccountSyncMessage(Message):
    """Syncing an account form master server v1 to v2."""

    accountid: Annotated[str, IOAttrs('id')]
    temp_name: Annotated[
        str | None, IOAttrs('temp_name', store_default=False)
    ] = None


@ioprepped
@dataclass
class PrivatePartyQueryMessage(Message):
    """A message from the v1 server forwarding a private-party-query."""

    account_id: Annotated[str, IOAttrs('a')]
    tickets: Annotated[int, IOAttrs('tk')]
    user_agent_string: Annotated[str, IOAttrs('u')]

    @override
    @classmethod
    def get_response_types(cls) -> list[type[Response] | None]:
        return [StringResponse]


@ioprepped
@dataclass
class PrivatePartyConnectMessage(Message):
    """A message from the v1 server forwarding a private-party-connect."""

    account_id: Annotated[str, IOAttrs('a')]
    code: Annotated[str, IOAttrs('c')]
    user_agent_string: Annotated[str, IOAttrs('u')]

    @override
    @classmethod
    def get_response_types(cls) -> list[type[Response] | None]:
        return [StringResponse]


@ioprepped
@dataclass
class PrivatePartyStartMessage(Message):
    """A message from the v1 server forwarding a private-party-start."""

    account_id: Annotated[str, IOAttrs('a')]
    tickets: Annotated[int, IOAttrs('tk')]
    config: Annotated[str, IOAttrs('c')]
    pings: Annotated[dict[str, float], IOAttrs('p')]
    user_agent_string: Annotated[str, IOAttrs('u')]

    @override
    @classmethod
    def get_response_types(cls) -> list[type[Response] | None]:
        return [StringResponse]


@ioprepped
@dataclass
class PrivatePartyStopMessage(Message):
    """A message from the v1 server forwarding a private-party-stop."""

    account_id: Annotated[str, IOAttrs('a')]
    tickets: Annotated[int, IOAttrs('tk')]
    user_agent_string: Annotated[str, IOAttrs('u')]

    @override
    @classmethod
    def get_response_types(cls) -> list[type[Response] | None]:
        return [StringResponse]


@ioprepped
@dataclass
class AdVerificationMessage(Message):
    """Ad ad-message-verification event from the v1 server."""

    reason: Annotated[str, IOAttrs('r')]
    account_id: Annotated[str, IOAttrs('a')]
    client_id: Annotated[str, IOAttrs('c')]
    t_id: Annotated[str, IOAttrs('tid')]
    build_number: Annotated[int, IOAttrs('b')]
    user_agent_string: Annotated[str, IOAttrs('u')]

    @override
    @classmethod
    def get_response_types(cls) -> list[type[Response] | None]:
        return [BoolResponse]


@ioprepped
@dataclass
class V1IAPVerifyMessage(Message):
    """Asks us to verify an IAP from the V1 side."""

    store: Annotated[V1Store, IOAttrs('s')]
    receipt: Annotated[str, IOAttrs('r')]
    item: Annotated[str, IOAttrs('i')]
    is_consumable: Annotated[bool, IOAttrs('c')]
    order_id: Annotated[str, IOAttrs('o')]
    accountid: Annotated[str, IOAttrs('a')]
    build_number: Annotated[int, IOAttrs('b')]
    user_agent_string: Annotated[str, IOAttrs('u')]
    ip_address: Annotated[str, IOAttrs('ip')]
    signature: Annotated[str | None, IOAttrs('sig', store_default=False)] = None

    @override
    @classmethod
    def get_response_types(cls) -> list[type[Response] | None]:
        return [BoolResponse]


@ioprepped
@dataclass
class SendMailMessage(Message):
    """Ask v2 to email someone."""

    address: Annotated[str, IOAttrs('a')]
    subject: Annotated[str, IOAttrs('s')]
    contents: Annotated[str, IOAttrs('c')]

    @override
    @classmethod
    def get_response_types(cls) -> list[type[Response] | None]:
        return [None]
