# Synced from ballistica-master-server.
# EFRO_SYNC_HASH=264630548149669643856970596517136705008
#
"""Small handy bits of functionality shared between server components."""

from __future__ import annotations

import asyncio
import subprocess
from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    pass


def current_process_username() -> str:
    """Return the currently signed in user."""
    import os
    import pwd

    return pwd.getpwuid(os.getuid())[0]


def run(cmd: str, log: Path | str | None = None) -> None:
    """Run a shell command and log its output.

    Throws exceptions for non-zero return values.
    If 'log' is provided, its contents will be read and echoed in
    this process with matching log-levels.
    """
    if log is not None:
        assert isinstance(log, (Path, str))
        raise RuntimeError('fixme not implemented yet.')
    results = subprocess.run(
        cmd, shell=True, stdin=subprocess.DEVNULL, check=False
    )
    #     cmd, shell=True, stdin=subprocess.DEVNULL, capture_output=True)
    # if results.stdout != b'':
    #     logging.info(results.stdout.decode())
    # if results.stderr != b'':
    #     logging.error(results.stderr.decode())
    if results.returncode != 0:
        raise RuntimeError(
            f"Command '{cmd}' returned non-zero"
            f' exit status {results.returncode}'
        )


async def async_run(cmd: str, log: Path | str | None = None) -> None:
    """Run a shell command asynchronously and log its output.

    Throws an exception on non-zero return values.
    If 'log' is provided, its contents will be read and echoed in
    this process with matching log-levels. Otherwise, final stdout
    is logged to 'info' and final stderr is logged to 'error'.
    """
    process = await asyncio.create_subprocess_shell(
        cmd,
        stdin=asyncio.subprocess.DEVNULL,
        # stdout=asyncio.subprocess.PIPE,
        # stderr=asyncio.subprocess.PIPE,
    )
    if log is not None:
        assert isinstance(log, (Path, str))
        raise RuntimeError('fixme not implemented yet.')
    # stdout, stderr = await process.communicate()
    # assert isinstance(stdout, bytes)
    # assert isinstance(stderr, bytes)
    await process.wait()
    # if stdout != b'':
    #     logging.info(stdout.decode())
    # if stderr != b'':
    #     logging.error(stderr.decode())
    if process.returncode != 0:
        raise RuntimeError(
            f"Command '{cmd}' returned non-zero"
            f' exit status {process.returncode}'
        )


def generate_self_signed_certificate(
    key_path: str,
    cert_path: str,
    ipaddress: str | None = None,
    common_name: str | None = None,
) -> None:
    """SSL goodness."""

    if (ipaddress is None) == (common_name is None):
        raise ValueError('must supply either ipaddress OR common_name')

    subj = (
        '/C=US'
        '/ST=California'
        '/L=San Francisco'
        '/O=Eric Froemling'
        '/OU=Eric Froemling'
    )
    subj += f'/CN={common_name}'

    # We don't expect these to stick around *too* long, but give enough
    # time that they can keep working for a while in case there's some
    # disaster where we can't spin up replacements immediately.
    days: int = 60

    # We want misconfigured devices to be able to communicate with us;
    # if we set start time to now then devices incorrectly set a short
    # amount of time in the past won't be able to talk to us. So let's
    # set start-time a bit in the past.
    predays: int = 10

    # We can use 'faketime -f -Xd CMD` to run CMD X days in the past.
    args: list[str] = [
        'faketime',
        '-f',
        f'-{predays}d',
        'openssl',
        'req',
        '-x509',
        '-newkey',
        'rsa:4096',
        '-keyout',
        key_path,
        '-out',
        cert_path,
        '-sha256',
        '-days',
        f'{days + predays}',
        '-nodes',
        '-subj',
        subj,
    ]
    if ipaddress is not None:
        args += ['-addext', f'subjectAltName = IP:{ipaddress}']

    subprocess.run(args, check=True)

    # Debugging: check start/end dates.
    if bool(False):
        print('CERT START DATE IS')
        subprocess.run(
            ['openssl', 'x509', '-startdate', '-noout', '-in', cert_path],
            check=False,
        )
        print('CERT END DATE IS')
        subprocess.run(
            ['openssl', 'x509', '-enddate', '-noout', '-in', cert_path],
            check=False,
        )
