"""Logging related functionality."""

from __future__ import annotations

# IMPORTANT: We may get pulled in by env so we can't import any other
# basn stuff here.

import os
import logging
import datetime
from typing import TYPE_CHECKING

from typing_extensions import override
from efro.log import setup_logging, LogLevel

from baserver.enums import ServerNodeRunMode

if TYPE_CHECKING:
    from typing import Any

    from efro.log import LogHandler

    from basn.taskwrapper import BaseTaskWrapper

loghandler: LogHandler

# Our top level logger. All our stuff is a child of this.
balog: logging.Logger
pingtargetlog: logging.Logger
transportlog: logging.Logger
hostedgamelog: logging.Logger
cloudbuildlog: logging.Logger


def setup_basn_logging() -> None:
    """Initial logging setup."""
    # pylint: disable=cyclic-import
    # pylint: disable=global-statement
    from basn.env import config

    # Set up logging defaults based on our config's ServerNodeRunMode
    # value.
    #
    #   RELEASE: INFO for root logger, WARNING for many unimportant ones.
    #   TEST: INFO for root logger, WARNING for few unimportant ones.
    #   DEBUG: INFO for root logger, DEBUG for ba logger.
    #
    # Let's also disable stderr log echoing for a bit of extra efficiency
    # (unless we're running manually in which case we want to see it).
    global loghandler
    loghandler = setup_logging(
        log_path=config.log_path,
        level=LogLevel.INFO,
        log_stdout_stderr=True,
        echo_to_stderr=(os.environ.get('BASN_MANUAL_RUN') == '1'),
        cache_size_limit=1024 * 100,  # ~100k of logs in mem
        cache_time_limit=datetime.timedelta(hours=1),
    )

    # Turn down logging levels on some things we don't need to hear.
    if config.run_mode.level < ServerNodeRunMode.DEBUG.level:
        for logname in ['aiohttp.access', 'asyncio']:
            logging.getLogger(logname).setLevel(logging.WARNING)

    class TidyLogFilter(logging.Filter):
        """Hide specific log messages."""

        @override
        def filter(self: Any, record: Any) -> bool:
            excstr = (
                str(record.exc_info[1]) if record.exc_info is not None else ''
            )

            # We sometimes get errors if broken requests comes in and
            # have no way to catch them. See
            # https://github.com/aio-libs/aiohttp/issues/5358
            if record.name == 'aiohttp.server' and record.levelname == 'ERROR':
                # Cases we know we want to hide.
                if record.getMessage() == 'Error handling request' and (
                    "Bad status line 'Invalid method encountered'" in excstr
                    or "400, message='Pause on PRI/Upgrade'" in excstr
                ):
                    return False  # Hide this record.

                # Show info on ones that get through in case we want to
                # hide them.
                dbgmsg = (
                    f'FILTER ALLOWING RECORD:'
                    f' {record.name=} {record.levelname=}'
                    f' {record.getMessage()=} {record=} {record.exc_info=}'
                    f' {excstr=}'
                )

                # Seeeems to be ok logging from within a log filter...
                balog.info(dbgmsg)

            return True

    logging.getLogger('aiohttp.server').addFilter(TidyLogFilter())

    # NOTE: Not suppressing this anymore; turns out it resets itself to
    # DEBUG when importing it so we'd have to do that first. Also our
    # suppress_non_root_debug filtering covers it fine currently.
    # logging.getLogger('filelock').setLevel(logging.WARNING)

    # Note: standard practice with Python loggers is to map loggers to
    # the module hierarchy. However, we instead organize things by a
    # specific set of categories, each under a single 'ba' logger. This
    # lets us tune the log set based on what is useful to see grouped
    # together and have individual controls over, and also allows us to
    # share log categories between clients and servers, which is useful
    # when viewing combined log streams.
    global balog
    balog = logging.getLogger('ba')

    # In our debug run level we turn debug on for this which will affect
    # all our loggers but not default system stuff like asyncio.
    #
    # NOTE TO SELF: Perhaps we should apply this to all our individual
    # sub-loggers which might lend itself better to manually changing
    # them at runtime in a predictable way.
    if config.run_mode.level >= ServerNodeRunMode.DEBUG.level:
        balog.setLevel(logging.DEBUG)

    global pingtargetlog
    pingtargetlog = logging.getLogger('ba.pingtarget')

    global transportlog
    transportlog = logging.getLogger('ba.transport')

    global cloudbuildlog
    cloudbuildlog = logging.getLogger('ba.cloudbuild')

    global hostedgamelog
    hostedgamelog = logging.getLogger('ba.hostedgame')


def logextra(
    channel: str | None = None,
    task: BaseTaskWrapper | str | None = None,
    session: str | None = None,
) -> dict:
    """Generate an 'extra' dict to include with log entries.

    This allows easy grouping of log messages in google cloud logging
    based on account-id, servernode-id, or various other things.

    A 'labels' dict will always exist in output; other entries may be
    created in the future if helpful.

    Note that some values may implicitly set others. For example,
    providing a ServerNode automatically sets Zone or providing a
    channel automatically sets task and servernode.

    For this reason, it is preferable to pass objects instead of ids as
    much as possible, as this allows more ancillary info to be added to
    the log automatically.
    """
    from basn.env import config

    # pylint: disable=cyclic-import
    labels: dict[str, str] = {}

    if channel is not None:
        splits = channel.split(':')
        if len(splits) != 3 or not all(splits):
            logging.error('Invalid channel passed to logextra: %s', channel)
        # Channel address is SERVERNODE:TASK:NUM; therefore we can pull
        # task from this if it wasn't explicitly provided.
        if task is None:
            task = ':'.join(splits[:2])
        labels['baAccountChannel'] = channel
    if task is not None:
        if not isinstance(task, str):
            task = (
                f'{config.node_id}:{task.task_id}'
                if task.has_task_id()
                else 'TASK_ID_NOT_YET_ASSIGNED'
            )
        else:
            # Task should be in string form SERVERNODE:TASK.
            splits = task.split(':')
            if len(splits) != 2 or not all(splits):
                logging.error('Invalid task passed to logextra: %s', task)
        labels['baBasnTask'] = task
    if session is not None:
        labels['baAccountSession'] = session

    return {'labels': labels}
