"""Wrangles shared things such as game binaries."""

from __future__ import annotations

import os
import time
import asyncio
import subprocess
from enum import Enum
from typing import TYPE_CHECKING, assert_never

from typing_extensions import override
from filelock import FileLock
from efro.dataclassio import IOMultiType

from basn.logging import balog, logextra

if TYPE_CHECKING:
    from basntask.command import TaskCommand


class ResourceTypeID(Enum):
    """Registry of resource types."""

    STANDARD_TASK_LIBS = 'standard_task_libs'
    SERVER_PACKAGE = 'server_package'
    CLOUD_FILE = 'cloud_file'


class Resource(IOMultiType[ResourceTypeID]):
    """A shared component usable by multiple tasks.

    Uses locks and permissions to ensure that creation attempts and
    access from multiple processes/threads is safe.
    """

    @override
    @classmethod
    def get_type_id(cls) -> ResourceTypeID:
        # Require child classes to supply this themselves. If we did a
        # full type registry/lookup here it would require us to import
        # everything and would prevent lazy loading.
        raise NotImplementedError()

    @override
    @classmethod
    def get_type(cls, type_id: ResourceTypeID) -> type[Resource]:
        # pylint: disable=cyclic-import
        out: type[Resource]
        t = ResourceTypeID

        if type_id is t.STANDARD_TASK_LIBS:
            from basn.resource.standardtasklibs import StandardTaskLibs

            out = StandardTaskLibs
        elif type_id is t.SERVER_PACKAGE:
            from basn.resource.serverpackage import ServerPackage

            out = ServerPackage
        elif type_id is t.CLOUD_FILE:
            from basn.resource.cloudfile import CloudFile

            out = CloudFile
        else:
            # Important to make sure we provide all types.
            assert_never(type_id)
        return out

    def get_name(self) -> str:
        """Return a globally unique name for this resource."""
        raise RuntimeError('This should be overridden.')

    async def prepare_async(self) -> None:
        """Async version of prepare."""

        # By default, just run old synchronous prepare.
        await asyncio.get_running_loop().run_in_executor(None, self.prepare)

    def prepare(self) -> None:
        """Download/create the resource if needed."""

        from basn.env import app

        resources_root = app.resources_root
        name = self.get_name()
        resdir = os.path.join(resources_root, name)
        resdirtmp = f'{resdir}.tmp'
        reslockfile = os.path.join(resources_root, f'.locks/{name}')

        # Quick-out: If our dir name exists, it means it has been
        # successfully created already and we can use it as-is.
        if os.path.exists(resdir):
            assert os.path.isdir(resdir)
            balog.info("Using existing resource: '%s'.", name, extra=logextra())
        else:
            # Lock to ensure we're the only ones writing to the
            # temp/real dir.
            balog.info("Creating resource '%s'...", name, extra=logextra())
            starttime = time.time()

            with FileLock(reslockfile):
                # There's a chance that someone else created the
                # resource while we were waiting for the lock.
                if os.path.exists(resdir):
                    assert os.path.isdir(resdir)
                    balog.info(
                        "Using existing resource: '%s'.", name, extra=logextra()
                    )
                else:
                    # We've got the lock and resource doesn't exist; its
                    # on us to create it.
                    subprocess.run(['rm', '-rf', resdirtmp], check=True)
                    os.mkdir(resdirtmp)

                    # Let the resource create itself.
                    self.create(resdirtmp)

                    # Set default permissions for dirs and files; we
                    # want them accessible but read only by any user.
                    #
                    # NOTE: these error on empty dirs, so make sure to
                    #       check.
                    if os.listdir(resdirtmp):
                        subprocess.run(
                            f'find {resdirtmp} -type d -print0 '
                            f'| xargs -0 chmod 755',
                            shell=True,
                            check=True,
                        )
                        subprocess.run(
                            f'find {resdirtmp} -type f -print0 '
                            f'| xargs -0 chmod 644',
                            shell=True,
                            check=True,
                        )
                    self.postcreate(resdirtmp)

                    os.rename(resdirtmp, resdir)

            balog.info(
                "Created resource '%s' in %.1f seconds.",
                name,
                time.time() - starttime,
                extra=logextra(),
            )

    def create(self, dirpath: str) -> None:
        """Create the resource in the given directory."""

    def postcreate(self, dirpath: str) -> None:
        """Called after default file permissions are set; for overriding."""

    def get_python_paths(self) -> list[str]:
        """Return any Python paths provided by this resource."""
        return []

    def get_ba_server_path(self) -> str | None:
        """If this resource provides a server, give its path."""
        return None
