"""Wrangles shared things such as game binaries."""

from __future__ import annotations

import os
from dataclasses import dataclass
from typing import TYPE_CHECKING, Annotated

from typing_extensions import override
from efro.dataclassio import ioprepped, IOAttrs

from basn.resource import Resource, ResourceTypeID


if TYPE_CHECKING:
    pass


@ioprepped
@dataclass
class CloudFile(Resource):
    """A cloud file."""

    cloud_file_id: Annotated[str, IOAttrs('f')]

    @override
    @classmethod
    def get_type_id(cls) -> ResourceTypeID:
        return ResourceTypeID.SERVER_PACKAGE

    @override
    def get_name(self) -> str:
        return f'cloudfile_{self.cloud_file_id}'

    def get_local_path(self) -> str:
        """Return the path where our data (will) live."""
        from basn.env import app

        return os.path.join(app.resources_root, self.get_name(), 'data')

    @override
    def create(self, dirpath: str) -> None:
        import urllib.request
        from baserver.cloudfile import cloud_file_id_for_path

        from basn.env import app, config

        # Note: We run in a synchronous worker thread so we use regular
        # non-async calls here.

        path = os.path.join(dirpath, 'data')

        url = (
            f'{config.master_server_addr}/'
            f'basn_cloudfile_get?f={self.cloud_file_id}'
            f'&n={config.node_id}&ct={app.comm_token}'
        )

        # Stream it straight disk in case its big.
        bytes_written = 0
        with (
            urllib.request.urlopen(
                url, context=app.default_ssl_context
            ) as response,
            open(path, 'wb') as out_file,
        ):
            chunk_size = 8192
            while True:
                chunk = response.read(chunk_size)
                if not chunk:
                    break
                out_file.write(chunk)
                bytes_written += len(chunk)

        # Make sure our data matches its id.
        checkid = cloud_file_id_for_path(path)

        if checkid != self.cloud_file_id:
            raise RuntimeError(
                'Downloaded cloud-file data does not match its id hash.'
            )

        # Success!
