"""Wrangles shared things such as game binaries."""

from __future__ import annotations

import os
import subprocess
from enum import Enum
from dataclasses import dataclass
from typing import TYPE_CHECKING, assert_never, Annotated

from typing_extensions import override
from efro.dataclassio import ioprepped, IOAttrs

from basn.resource import Resource, ResourceTypeID


if TYPE_CHECKING:
    pass


@ioprepped
@dataclass
class ServerPackage(Resource):
    """A particular version of the ba server."""

    class Flavor(Enum):
        """Type of server."""

        VANILLA = 'vanilla'
        EON = 'eon'

    flavor: Annotated[Flavor, IOAttrs('flavor', soft_default=Flavor.VANILLA)]

    build_number: int

    @override
    @classmethod
    def get_type_id(cls) -> ResourceTypeID:
        return ResourceTypeID.SERVER_PACKAGE

    @override
    def get_name(self) -> str:
        if self.flavor is self.Flavor.VANILLA:
            suffix = ''
        elif self.flavor is self.Flavor.EON:
            suffix = '_eon'
        else:
            assert_never(self.flavor)
        return f'server_package{suffix}_{self.build_number}'

    @override
    def create(self, dirpath: str) -> None:
        import platform

        assert platform.system() == 'Linux'
        machine = platform.machine()
        if machine == 'x86_64':
            basename = 'x86_64'
        elif machine == 'aarch64':
            basename = 'arm64'
        else:
            raise RuntimeError(f'invalid machine type {machine}')
        archivename = f'{basename}.tar.gz'

        # Determine where we pull this from.
        if self.flavor is self.Flavor.VANILLA:
            payload_url = (
                f'https://files.ballistica.net/test/bab'
                f'/{self.build_number}/{archivename}'
            )
            strip = ' && rm baserver/README.txt baserver/config.yaml'
        elif self.flavor is self.Flavor.EON:
            payload_url = (
                f'https://files.ballistica.net/test/eon'
                f'/{self.build_number}/{archivename}'
            )
            # We use the config as a starting point in this case.
            strip = ''
        else:
            assert_never(self.flavor)

        # Kill any dirs starting with '.' - am seeing resource fork
        # stuff here sometimes (._somefile.py) which is causing problems.
        strip2 = ' && find baserver -name ".*" -delete'
        subprocess.run(
            f'cd {dirpath}'
            f' && wget --quiet {payload_url}'
            f' && tar -xf {archivename} && rm {archivename}'
            f' && mv {basename} baserver'
            f'{strip}{strip2}',
            shell=True,
            check=True,
        )

    def getappname(self) -> str:
        """Base name for the dist we're using."""
        name: str
        if self.flavor is self.Flavor.VANILLA:
            name = 'ballisticakit'
        elif self.flavor is self.Flavor.EON:
            name = 'bombsquad'
        else:
            assert_never(self.flavor)
        return name

    @override
    def postcreate(self, dirpath: str) -> None:
        appname = self.getappname()

        # Our binaries need to be executable (and for some reason aren't
        # when we decompress them).
        subprocess.run(
            [
                'chmod',
                '+x',
                os.path.join(dirpath, f'baserver/{appname}_server'),
                os.path.join(dirpath, f'baserver/dist/{appname}_headless'),
            ],
            check=True,
        )

    @override
    def get_ba_server_path(self) -> str | None:
        from basn.env import app

        appname = self.getappname()
        return os.path.join(
            app.resources_root, self.get_name(), 'baserver', f'{appname}_server'
        )
