"""Functionality for wrangling tasks."""

from __future__ import annotations

import random
import datetime
from collections import deque
from typing import TYPE_CHECKING, assert_type

from efro.util import utc_now
from efro.log import LogLevel
from baserver.msg import bamastertobasn
from baserver.data import BASNStatsSparse
from baserver.enums import ServerNodeTaskState

from basn.logging import logextra
from basn.taskwrapper import InlineTaskWrapper

if TYPE_CHECKING:
    from typing import Dict, Set

    from baserver.enums import TaskTestType

    from basn.taskwrapper import BaseTaskWrapper


class TaskNotFoundError(Exception):
    """Error thrown when requesting a dead task."""


class TaskSet:
    """Wrangles a set of running tasks."""

    def __init__(self) -> None:
        from basn.env import config

        self._next_task_id = 0
        self._tasks: Dict[int, BaseTaskWrapper] = {}

        # Create a shuffled list of udp_ports we can use.
        # We'll pop from the front and push to the back so that
        # we don't reuse a port too quickly (minimizes chance of
        # random chatter from one game bleeding into the next)
        self._free_udp_ports = deque[int](
            range(config.udp_port_min, config.udp_port_max + 1)
        )
        random.shuffle(self._free_udp_ports)

        # Do the same with users for the same reason
        # (want to minimize the chance that killing all a user's tasks
        # could affect the next user due to a race condition, etc.)
        assert config.num_users > 0
        self._free_users = deque[str](
            f'ba{i:03d}' for i in range(config.num_users)
        )
        random.shuffle(self._free_users)

    def app_did_start_running(self) -> None:
        """Called when the containing app is done starting up."""
        self.update_app_running_and_load_stats()

    def reset_errors(self) -> None:
        """Reset all task error flags."""
        from basn.env import app

        app.stats.tasks_errored = False
        app.stats.max_task_log_level = LogLevel.INFO
        for task in self._tasks.values():
            task.reset_errors()

    async def minute_upkeep(self) -> None:
        """Perform various upkeep."""
        for task in self._tasks.values():
            if task.state is ServerNodeTaskState.RUNNING:
                await task.minute_upkeep()

        # Prune cleanly-ended or killed tasks that had no warnings/errors
        # after a while.
        keep_tasks: dict[int, BaseTaskWrapper] = {}
        for task_id, task in sorted(self._tasks.items()):
            if (
                task.state
                in {ServerNodeTaskState.COMPLETED, ServerNodeTaskState.KILLED}
                and task.max_log_level.value <= LogLevel.INFO.value
                and not task.error_flag
            ):
                assert task.duration is not None
                endtime = task.launch_time + datetime.timedelta(
                    seconds=task.duration
                )
                since_ended = utc_now() - endtime
                if since_ended >= datetime.timedelta(minutes=1.0):
                    task.logger.debug(
                        'Pruning old task %d: %s',
                        task_id,
                        task,
                        extra=logextra(task=task),
                    )
                    continue
            keep_tasks[task_id] = task
        self._tasks = keep_tasks

    def task_by_id(self, task_id: int) -> BaseTaskWrapper:
        """Given a numeric id, return a task."""
        try:
            return self._tasks[task_id]
        except KeyError:
            raise TaskNotFoundError from None

    def launch_task(self, task: BaseTaskWrapper) -> int:
        """Attempt to launch a newly created task.

        Raises CleanError for any cleanly-handled reason the task can't
        be launched. Other exceptions should be considered errors that
        should be fixed.
        """
        # pylint: disable=cyclic-import
        from efro.error import CleanError

        from basn.env import app
        from basn.app import State

        # If we don't yet have a hard-capacity or speed factor set,
        # we can't launch anything because we don't know our limits.
        if app.stats.hard_capacity is None:
            raise CleanError("Can't launch task without hard_capacity set.")

        # Make sure the app is in a reasonable state for this.
        # We can launch tasks while running, and also while waiting to
        # shut down (we may be needed for emergency capacity).
        if app.state not in {State.RUNNING, State.SHUTTING_DOWN}:
            raise CleanError(f"Can't launch task in app-state {app.state}.")

        # Make sure we've got sufficient load capacity for this.
        # Generally bamaster should have a handle on this and avoid sending
        # us more load than we can handle, but if there's a rush of task
        # launches its possible that bamaster may be looking at out-of-date
        # stats for us. So we have the final say here.
        assert app.stats.load is not None
        if app.stats.load + task.load > app.stats.hard_capacity:
            raise CleanError(
                f'Insufficient load capacity to launch task '
                f'({task.load:.2f} required, '
                f'{app.stats.hard_capacity - app.stats.load:.2f}'
                f' available)'
            )

        # If we've got too many sockets open, reject all new tasks.
        # By default linux can handle 1000 open sockets at once.
        # (we should try to calibrate task load values so we don't get
        # to this point).
        if app.sockets_danger:
            raise CleanError(
                f'Too many sockets open ({app.stats.open_sockets_count}) '
                f'to launch tasks.'
            )

        # Our mem-load calculations should prevent this from happening,
        # but if somehow our machine is getting dangerously low on memory,
        # refuse to launch stuff.
        if app.mem_danger:
            raise CleanError(
                "System memory is dangerously low; can't launch task."
            )

        # Find a free user-id if one is needed.
        user_id: str | None = None
        if task.get_needs_user_id():
            if not self._free_users:
                task.logger.critical(
                    'No free user_ids available.',
                    extra=logextra(task=task),
                )
                raise CleanError('No free user_ids available.')
            user_id = self._free_users.popleft()

        # If they need a udp_port, find a free one.
        udp_port: int | None = None
        if task.get_needs_udp_port():
            if not self._free_udp_ports:
                task.logger.critical(
                    'No free udp_ports available.',
                    extra=logextra(task=task),
                )
                raise CleanError('No free udp_ports available.')
            udp_port = self._free_udp_ports.popleft()

        # Make sure to use unique task nums even if errors occur
        # (don't want to trip over partially constructed prior attempt)
        task_id = self._next_task_id
        self._next_task_id += 1
        self._tasks[task_id] = task

        # Update our load stats immediately before we actually kick off
        # the task. This gets called again once the task reaches the
        # running state, but that might not happen for a moment and we
        # want to keep load up to date so we can't exceed our hard cap
        # in the meantime.
        self.update_app_running_and_load_stats()

        task.launch(self, user_id, task_id, udp_port)
        return task_id

    def task_is_running(self, task: BaseTaskWrapper) -> None:
        """Called when a task is up and running."""
        del task  # Unused.
        # Update our total load stats to account for this task running.
        self.update_app_running_and_load_stats()

    def task_errored(self, task: BaseTaskWrapper) -> None:
        """Called by a task if it errored."""
        from basn.env import app

        del task  # Unused.
        app.stats.tasks_errored = True
        app.mark_stats_dirty()

    def task_finished(self, task: BaseTaskWrapper) -> None:
        """Called by a task when its done, regardless of outcome."""
        from basn.util import force_kill_user_processes

        if task.user_id is not None:
            # Kill any processes running under this user.
            # (though we should try to never have any at this point).
            force_kill_user_processes(
                user_id=task.user_id, for_task=task, warn_if_running=True
            )

            task.logger.debug(
                'Task %s finished; releasing user %s.',
                task.task_id,
                task.user_id,
                extra=logextra(task=task),
            )
            # Return our user back to the pool.
            self._free_users.append(task.user_id)
        else:
            task.logger.debug(
                'Task %s finished.',
                task.task_id,
                extra=logextra(task=task),
            )

        # If we were using a udp_port, add it back to the pool.
        # (we should be 100% certain that the task processes using it are
        # dead at this point).
        if task.udp_port is not None:
            self._free_udp_ports.append(task.udp_port)

        # Update our total load stats to account for this task dying.
        self.update_app_running_and_load_stats()

    def update_app_running_and_load_stats(self) -> None:
        """Update our total current app load."""
        from basn.env import app

        # Tally totals.
        load = 0.0
        running_inline = 0
        running_standalone = 0
        for _task_id, task in sorted(self._tasks.items()):

            # We update load as soon as tasks are launched; otherwise
            # it is possible to exceed our hard cap by running launches
            # in a loop.
            if (
                task.state is ServerNodeTaskState.LAUNCHING
                or task.state is ServerNodeTaskState.RUNNING
            ):
                load += task.load

            # Don't count it as a running task until it is actually
            # running through.
            if task.state is ServerNodeTaskState.RUNNING:
                if isinstance(task, InlineTaskWrapper):
                    running_inline += 1
                else:
                    running_standalone += 1

        # Apply to stats and mark dirty ONLY if something differs.
        if (
            app.stats.load != load
            or app.stats.running_task_count_inline != running_inline
            or app.stats.running_task_count_standalone != running_standalone
        ):
            app.stats.load = load
            app.stats.running_task_count_inline = running_inline
            app.stats.running_task_count_standalone = running_standalone
            app.mark_stats_dirty()

    def get_sparse_stats(self, stats: BASNStatsSparse) -> None:
        """Provide sparse stats in the provided stats obj."""
        from basn.taskwrapper.transportagent import TransportAgentTaskWrapper

        for task_id, task in sorted(self._tasks.items()):
            if task.state is ServerNodeTaskState.RUNNING:
                if isinstance(task, TransportAgentTaskWrapper):
                    cstats = BASNStatsSparse.TransportClientStats(
                        session_channels=task.get_channel_ids_per_session(
                            ready_only=True
                        )
                    )
                    stats.transport_clients[task_id] = cstats

    def invalidate_token(self, token: str) -> None:
        """Invalidate a session token from bamaster."""
        from basn.taskwrapper.transportagent import TransportAgentTaskWrapper

        for task in self._tasks.values():
            if task.state is ServerNodeTaskState.RUNNING:
                if isinstance(task, TransportAgentTaskWrapper):
                    task.invalidate_token_on_client(token)

    def get_tasks_info_message(
        self, min_log_level: LogLevel
    ) -> bamastertobasn.TasksQueryResponse:
        """Construct a message describing our current set of tasks."""

        msg = bamastertobasn.TasksQueryResponse()
        for task_id, task in sorted(self._tasks.items()):
            if min_log_level.value > 0:
                # Dealing in enum values but want to make sure the
                # actual types still match.
                assert_type(task.max_log_level, LogLevel)

                # Calc a single log-level for the task which
                # also accounts for whether it errored or not.
                task_log_level = task.max_log_level
                if (
                    task_log_level.value < LogLevel.ERROR.value
                    and task.error_flag
                ):
                    task_log_level = LogLevel.ERROR

                if min_log_level.value > task_log_level.value:
                    continue

            msg.entries.append(
                bamastertobasn.TasksInfoMessageEntry(
                    task_id=task_id,
                    label=task.label,
                    user_id=task.user_id,
                    launch_time=task.launch_time,
                    duration=task.duration,
                    state=task.state,
                    load=task.load,
                    error_flag=task.error_flag,
                    max_log_level=task.max_log_level,
                    actions=task.get_actions(),
                    udp_port=task.udp_port,
                )
            )
        return msg
