"""Task class for test task."""

from __future__ import annotations

import copy
import datetime
from typing import TYPE_CHECKING, assert_never

from typing_extensions import override
from baserver.msg import bamastertobasn
from baserver.hostedgame import HostedGameType, AdHocGameType
from baserver.enums import ServerNodeTaskState

from basn.taskwrapper import StandaloneTaskWrapper
from basn.logging import hostedgamelog

if TYPE_CHECKING:
    from basntask.command import TaskCommand
    from baserver.hostedgame import HostedGameDef

    from basn.resource import Resource

CONNECT_TO_GAME_ACTION = 'Connect to Game'


class HostedGameTaskWrapper(StandaloneTaskWrapper):
    """A task which hosts a game."""

    def __init__(self, gamedef: HostedGameDef) -> None:
        super().__init__(logger=hostedgamelog)

        self.gamedef = copy.deepcopy(gamedef)

    @override
    def get_process_priority(self) -> StandaloneTaskWrapper.ProcessPriority:
        # We need to be real-time-ish.
        return StandaloneTaskWrapper.ProcessPriority.HIGH

    @override
    def get_load(self) -> float:
        # bamaster informs us how heavy this task will be.
        return self.gamedef.load

    @override
    def get_needs_udp_port(self) -> bool:
        return True

    @override
    def get_label(self) -> str:
        from efro.dataclassio import dataclass_from_json
        from bacommon.net import PrivateHostingConfig

        if self.gamedef.hosted_game_type is HostedGameType.STRESS_TEST:
            return f'stress-test-{self.gamedef.stress_test_minutes:.0f}m'
        if self.gamedef.hosted_game_type is HostedGameType.AD_HOC:
            if self.gamedef.ad_hoc_game_type is AdHocGameType.EON:
                return 'ad-hoc: eon'
            ad_hoc_extra = (
                ''
                if self.gamedef.ad_hoc_name is None
                else (
                    ': teams'
                    if 'Teams' in self.gamedef.ad_hoc_name
                    else ': ffa' if 'FFA' in self.gamedef.ad_hoc_name else ''
                )
            )
            return f'ad-hoc{ad_hoc_extra}'
        if self.gamedef.hosted_game_type is HostedGameType.PRIVATE_PARTY:
            assert self.gamedef.private_hosting_config is not None
            hconfig = dataclass_from_json(
                PrivateHostingConfig, self.gamedef.private_hosting_config
            )
            return f'priv-party: {hconfig.session_type}'
        assert_never(self.gamedef.hosted_game_type)
        return 'Unknown HostedGameType'  # type: ignore

    @override
    def get_command(self) -> TaskCommand:
        from basntask.hostedgame import HostedGameCommand

        assert self.udp_port is not None
        return HostedGameCommand(self.gamedef, port=self.udp_port)

    @override
    def get_max_duration(self) -> datetime.timedelta:
        # For game types with hard limits, we force-kill everything if
        # things somehow survive past the limit.
        if self.gamedef.hosted_game_type is HostedGameType.STRESS_TEST:
            return datetime.timedelta(
                minutes=self.gamedef.stress_test_minutes + 10
            )
        if self.gamedef.hosted_game_type is HostedGameType.AD_HOC:
            return datetime.timedelta(
                minutes=self.gamedef.ad_hoc_minutes_unclean + 10
            )
        return super().get_max_duration()

    @override
    def get_resources(self) -> list[Resource]:
        from basn.env import DEFAULT_BALLISTICAKIT_VERSION, DEFAULT_EON_VERSION
        from basn.resource.standardtasklibs import StandardTaskLibs
        from basn.resource.serverpackage import ServerPackage

        return [
            StandardTaskLibs(),
            (
                ServerPackage(
                    flavor=ServerPackage.Flavor.EON,
                    build_number=DEFAULT_EON_VERSION,
                )
                if (
                    self.gamedef.hosted_game_type is HostedGameType.AD_HOC
                    and self.gamedef.ad_hoc_game_type is AdHocGameType.EON
                )
                else ServerPackage(
                    flavor=ServerPackage.Flavor.VANILLA,
                    build_number=DEFAULT_BALLISTICAKIT_VERSION,
                )
            ),
        ]

    @override
    def get_actions(self) -> list[str]:
        our_actions: list[str] = (
            [CONNECT_TO_GAME_ACTION]
            if self.state is ServerNodeTaskState.RUNNING
            else []
        )
        return our_actions + super().get_actions()

    @override
    def action(self, action: str) -> bamastertobasn.TaskActionResponse:
        assert self._root is not None
        assert ' ' not in str(self._root)
        if action == CONNECT_TO_GAME_ACTION:
            from basn.env import app
            import shlex

            pycmd = shlex.quote(
                f'import bascenev1; bascenev1.connect_to_party('
                f'address="{app.our_address}", port={self.udp_port})'
            )
            fullcmd = (
                f'cd ~/LocalDocs/ballistica-internal/build/cmake/debug/staged'
                f' && ./ballisticakit -e {pycmd}'
            )
            return bamastertobasn.TaskActionResponse(
                output=fullcmd, copy_text=fullcmd, auto_return=False
            )

        return super().action(action)
