"""Wrangle transport connection to a client."""

from __future__ import annotations

import time
import asyncio
import weakref
import datetime
from dataclasses import dataclass
from typing import TYPE_CHECKING, Awaitable

from typing_extensions import override
from efro.util import utc_now
from efro.rpc import RPCEndpoint, ssl_stream_writer_underlying_transport_info
from efro.call import tpartial
from efro.error import CommunicationError
import bacommon.cloud
from baserver.msg import batocloud
from baserver.msg import cloudtoba

from basn.taskwrapper import InlineTaskWrapper
from basn.taskwrapper.transportagent import _impl1
from basn.taskwrapper.transportagent import _impl2
from basn.mgen.clientsender import ClientMessageSender
from basn.mgen.clientreceiver import ClientMessageReceiver
from basn.logging import logextra, transportlog

if TYPE_CHECKING:
    from threading import Thread

    from efro.message import Message, Response, SysResponse
    from baserver.location import Location
    from baserver.enums import TaskTestType
    from basntask.command import TaskCommand

# DEBUG_PRINT = False
# LABEL = 'transportagent'


@dataclass
class AccountSessionChannel:
    """Channel to an account in use by a TransportAgent."""

    token: str

    # Whether this channel is ready for use.
    # (has been verified with bamaster and is still valid).
    ready: bool = False

    # Set when *anyone* has marked this channel as dead.
    closed: bool = False

    # Set when it has been confirmed with bamaster that this channel is dead.
    close_confirmed: bool = False

    # If bamaster has informed us that the session itself is dead.
    session_invalidated: bool = False

    info: batocloud.CredentialsCheckResponseInfo | None = None
    last_verify_attempt_time: float | None = None


class TransportAgentTaskWrapper(InlineTaskWrapper):
    """Task which sits and responds to pings."""

    _event_loop: asyncio.AbstractEventLoop

    # Communication with the connected client. These can be considered
    # reliable and will arrive in the order sent (If the connection goes
    # down, our whole task goes down).
    msg_out = ClientMessageSender()
    msg_in = ClientMessageReceiver()

    def __init__(
        self,
        connection_wait_seconds: float,
        max_duration_seconds: float,
        load: float,
        token: str,
        user_agent_string: str,
        remote_addr: str,
        location: Location | None,
    ) -> None:
        super().__init__(logger=transportlog)
        self._connection_wait_seconds = connection_wait_seconds
        self._max_duration = datetime.timedelta(seconds=max_duration_seconds)
        self._tload = load
        self._token = token
        self._endpoint: RPCEndpoint | None = None
        self._reader: asyncio.StreamReader | None = None
        self._writer: asyncio.StreamWriter | None = None
        self._client_connected = asyncio.Event()
        self._max_time_task: asyncio.Task | None = None
        self.tasks: list[asyncio.Task] = []
        self._ended = False

        # Override our log.
        self.logger = transportlog

        # Contains info the client sends us about themself.
        self._handshake: batocloud.TransportAgentHandshakeMessage | None = None

        # ids mapped to account-session-channels.
        self.session_channels: dict[int, AccountSessionChannel] = {}

        # Some info bamaster passes us about who is connecting.
        self.user_agent_string = user_agent_string
        self.remote_addr = remote_addr
        self.location = location

    @property
    def handshake(self) -> batocloud.TransportAgentHandshakeMessage:
        """Info the client sent us about itself."""
        if self._handshake is None:
            raise RuntimeError('handshake has not been set')
        return self._handshake

    @property
    def token(self) -> str:
        """Token our client uses to authenticate itself."""
        return self._token

    @override
    def get_needs_udp_port(self) -> bool:
        return False

    @override
    def get_load(self) -> float:
        # bamaster informs us how heavy this task will be.
        return self._tload

    @override
    def get_max_duration(self) -> datetime.timedelta:
        # We attempt to gracefully close our connection after some
        # amount of time. If we are somehow running a few minutes after
        # that then basn should drop the hammer.
        return self._max_duration + datetime.timedelta(minutes=5)

    def get_channel_ids_per_session(
        self, ready_only: bool
    ) -> dict[str, list[int]]:
        """Return session-channel-ids per session-id (token)."""

        assert self._handshake is None or self._handshake.ba_build >= 20888

        out: dict[str, list[int]] = {}
        for channel_id, handle in self.session_channels.items():
            if ready_only and not handle.ready:
                continue
            if handle.token not in out:
                out[handle.token] = []
            out[handle.token].append(channel_id)
        return out

    @override
    def get_label(self) -> str:
        return 'v2t-agent'

    @override
    async def run_inline(self) -> None:
        try:
            await self._do_run_inline()
        except asyncio.CancelledError:
            self.logger.warning(
                'TransportAgent %d got CancelledError; unexpected.',
                id(self),
                extra=logextra(task=self),
            )
            raise

    def _log_max_time_task_death(self) -> None:
        self.add_note('max_time_task dying')

    async def _do_run_inline(self) -> None:
        from basn.env import config

        self.add_note(f'Hello from transport-agent with token {self._token}.')

        self._event_loop = asyncio.get_running_loop()

        # Wait until someone signals us that we have streams available
        # (or have timed out).
        self.logger.debug('Waiting for client connection...')
        try:
            await asyncio.wait_for(
                self._client_connected.wait(),
                timeout=self._connection_wait_seconds,
            )
        except asyncio.TimeoutError:
            self.logger.info(
                'TransportAgent task %s:%s: timed out waiting'
                ' for client to connect.',
                config.node_id,
                self.task_id,
                extra=logextra(task=self),
            )
            self.add_note('Timed out waiting for client to connect.')
            return

        assert self._reader is not None and self._writer is not None

        # Create the endpoint and spit connection debug messages to our notes.
        self._endpoint = RPCEndpoint(
            tpartial(_impl1.handle_raw_message, self),
            self._reader,
            self._writer,
            debug_print=False,
            debug_print_io=False,
            debug_print_call=tpartial(_impl1.note_if_alive, weakref.ref(self)),
            label='transportagent',
        )

        # Kill the connection after max-time.
        # (could use asyncio.wait_for() here but would need to make sure
        # canceling endpoint.run() behaves correctly).
        self._max_time_task = asyncio.create_task(
            self._max_time_disconnect(),
            name=f'transportagent {self._task_id} max time disconnect',
        )
        try:
            run_start_time = time.monotonic()
            await self._endpoint.run()
            duration = time.monotonic() - run_start_time
            self.add_note(f'Endpoint run exited cleanly after {duration:.2f}s.')
        except Exception:
            self.logger.exception(
                'Error running transport rpc endpoint.',
                extra=logextra(task=self),
            )
        except BaseException as exc:
            self.add_note(f'BaseExc {type(exc)} coming out of run.')

        self._ended = True

        try:
            self._max_time_task.cancel()
        except Exception:
            self.logger.exception(
                'Error canceling max-time-task.',
                extra=logextra(task=self),
            )
        self._max_time_task = None

        # Close any remaining open channels and inform bamaster of their
        # demise.
        close_channel_ids: list[int] = []
        for channel_id, channel in self.session_channels.items():
            if not channel.closed:
                channel_addr = f'{config.node_id}:{self.task_id}:{channel_id}'
                self.logger.info(
                    'Closing channel %s of session %s at task shutdown.',
                    channel_addr,
                    channel.token,
                    extra=logextra(
                        task=self,
                        channel=channel_addr,
                        session=channel.token,
                    ),
                )
                channel.ready = False
                channel.closed = True
                close_channel_ids.append(channel_id)
        self.tasks.append(
            asyncio.create_task(
                _impl1.deregister_channels_with_bamaster(
                    self, close_channel_ids
                ),
                name=f'transportagent {self.task_id}'
                f' channel {close_channel_ids} deregister',
            )
        )

        # Wait for all tasks we made to end and report any errors they had.
        results = await asyncio.gather(*self.tasks, return_exceptions=True)
        for result in results:
            if isinstance(result, BaseException):
                self.logger.warning(
                    'TransportAgent %s:%s: error in subtask.',
                    config.node_id,
                    self._task_id,
                    exc_info=result,
                    extra=logextra(task=self),
                )

        # Clear them to hopefully break any dependency loops.
        self.tasks.clear()

        # Sanity check: all our channels should be confirmed closed.
        for channel_id, channel in self.session_channels.items():
            if (
                channel.ready
                or not channel.closed
                or not channel.close_confirmed
            ):
                channel_addr = f'{config.node_id}:{self.task_id}:{channel_id}'
                self.logger.warning(
                    'Found not-cleaned-up channel %s at task shutdown;'
                    ' ready=%s, closed=%s, close_confirmed=%s.',
                    channel_addr,
                    channel.ready,
                    channel.closed,
                    channel.close_confirmed,
                    extra=logextra(
                        task=self,
                        channel=channel_addr,
                        session=channel.token,
                    ),
                )

        self.logger.info(
            'TransportAgent %s:%s: task ending.',
            config.node_id,
            self.task_id,
            extra=logextra(task=self),
        )

    def is_ended(self) -> bool:
        """Have we ended?"""
        return self._ended

    async def client_connection_handoff(
        self,
        reader: asyncio.StreamReader,
        writer: asyncio.StreamWriter,
        handshake: batocloud.TransportAgentHandshakeMessage,
    ) -> None:
        """Hand over streams for a newly connected client.

        At this point the streams are owned by us.
        """
        from basn.env import config

        self._reader = reader
        self._writer = writer

        self.logger.info(
            'TransportAgent %s:%s: was assigned transport %s.',
            config.node_id,
            self.task_id,
            ssl_stream_writer_underlying_transport_info(self._writer),
            extra=logextra(task=self),
        )

        # Store the handshake which contains useful stuff like their
        # internal device id.
        self._handshake = handshake

        peername = writer.get_extra_info('peername')
        self.add_note(f'Connection established: {peername=} {handshake=}.')

        self._client_connected.set()

    # Handle max-time disconnect.
    async def _max_time_disconnect(self) -> None:
        warn_seconds = 30.0  # How long before shutdown we warn them.
        assert self._max_duration.total_seconds() > warn_seconds

        sleepsecs = self._max_duration.total_seconds() - warn_seconds
        self.add_note(f'Sleeping for {int(sleepsecs)} seconds...')

        # Note: can put this back to a single sleep call.
        if bool(True):
            await asyncio.sleep(sleepsecs)
        else:
            # Debugging - print progress periodically...
            assert self._launch_time is not None
            while True:
                await asyncio.sleep(61.23)
                since_start = (utc_now() - self._launch_time).total_seconds()
                if since_start > sleepsecs:
                    break
                self.add_note(f'Still sleeping {since_start:.1f}s in...')

        # Issue a warning to the client that the agent is about to go
        # down so they can start setting up a replacement and avoid
        # downtime.

        if self.handshake.ba_build >= 20888:
            self.add_note('Issuing shutdown warning...')

            try:
                await self.msg_out.send_async(
                    cloudtoba.TransportAgentShutdownWarningMessage()
                )
            except CommunicationError:
                self.add_note('Comm err issuing shutdown warning')
            except Exception:
                # We ignore comm errors since this is a best-effort situation,
                # but make noise if we see something we don't expect.
                self.logger.warning(
                    'Error sending transport-agent shutdown warning.',
                    exc_info=True,
                    extra=logextra(task=self),
                )
            except BaseException as exc:
                self.add_note(
                    f'Got {type(exc)} while sending shutdown warning.'
                )
                raise

        self.add_note(
            f'Sleeping for warn-seconds {warn_seconds:.0f} before close...'
        )
        await asyncio.sleep(warn_seconds)
        self.add_note('Will kick off max-time close...')

        # Note: create a task to do this instead of doing it inline here
        # since we don't want it to be possible to be cancelled in the
        # middle of the procedure.
        self.tasks.append(
            asyncio.create_task(
                self._do_max_time_close(),
                name=f'transportagent {self._task_id} max time close',
            )
        )

    async def _do_max_time_close(self) -> None:
        try:
            self.add_note(
                f'Max-time {self._max_duration} expired;'
                f' closing connection.'
            )
            if self._endpoint is not None:
                self.logger.info(
                    'Doing max-time disconnect for task %d.',
                    self.task_id,
                    extra=logextra(task=self),
                )
                self._endpoint.close()
                await self._endpoint.wait_closed()
            else:
                self.logger.error(
                    'Would do max-time disconnect but endpoint is None.',
                    extra=logextra(task=self),
                )
        except asyncio.CancelledError:
            self.logger.warning(
                'Got CancelledError for _do_max_time_close(); unexpected.',
                extra=logextra(task=self),
            )
            raise

    def invalidate_token_on_client(self, token: str) -> None:
        """Invalidate a token if we are using it (asynchronously)."""
        _impl1.invalidate_token_on_client(self, token)

    @msg_out.send_async_method
    def send_raw_message_async(self, data: str) -> Awaitable[str]:
        """Send a raw message to the client."""
        assert self._endpoint is not None

        # Do the send call immediately so we can guarantee send order,
        # and then wait for completion asynchronously.
        awaitable = self._endpoint.send_message(data.encode())
        return self._send_raw_message_async(awaitable)

    async def _send_raw_message_async(self, awaitable: Awaitable[bytes]) -> str:
        """Send a raw message to the client."""
        return (await awaitable).decode()

    @msg_in.decode_filter_method
    def _msg_in_decode_filter(self, indata: dict, message: Message) -> None:
        return _impl1.msg_in_decode_filter(self, indata, message)

    @msg_in.encode_filter_method
    def _msg_in_encode_filter(
        self,
        message: Message | None,
        response: Response | SysResponse,
        outdata: dict,
    ) -> None:
        return _impl1.msg_in_encode_filter(self, message, response, outdata)

    # Testing a message from the client we handle ourself.
    @msg_in.handler
    async def _handle_msg_1(
        self, msg: batocloud.CToBMsg1
    ) -> batocloud.CToBResp1:
        return await _impl1.handle_msg_1(self, msg)

    # Testing another message from the client that we handle ourself.
    @msg_in.handler
    async def _handle_msg_2(
        self, msg: batocloud.CToBMsg2
    ) -> batocloud.CToBResp1 | batocloud.CToBResp2:
        return await _impl2.handle_msg_2(self, msg)

    @msg_in.handler
    async def _handle_sign_in_message(
        self, msg: bacommon.cloud.SignInMessage
    ) -> bacommon.cloud.SignInResponse:
        return await _impl1.handle_sign_in_message(self, msg)

    @msg_in.handler
    async def _handle_login_proxy_request_message(
        self, msg: bacommon.cloud.LoginProxyRequestMessage
    ) -> bacommon.cloud.LoginProxyRequestResponse:
        return await _impl1.handle_login_proxy_request_message(self, msg)

    @msg_in.handler
    async def _handle_login_proxy_state_query_message(
        self, msg: bacommon.cloud.LoginProxyStateQueryMessage
    ) -> bacommon.cloud.LoginProxyStateQueryResponse:
        return await _impl1.handle_login_proxy_state_query_message(self, msg)

    @msg_in.handler
    async def _handle_login_proxy_complete_message(
        self, msg: bacommon.cloud.LoginProxyCompleteMessage
    ) -> None:
        return await _impl1.handle_login_proxy_complete_message(self, msg)

    @msg_in.handler
    async def _handle_account_session_release_message(
        self, msg: batocloud.AccountSessionReleaseMessage
    ) -> None:
        return await _impl1.handle_account_session_release_message(self, msg)

    @msg_in.handler
    async def _handle_credentials_check_message(
        self, msg: batocloud.CredentialsCheckMessage
    ) -> batocloud.CredentialsCheckResponse:
        return await _impl1.handle_credentials_check_message(self, msg)

    @msg_in.handler
    def _handle_account_channel_create_message(
        self, msg: batocloud.AccountChannelCreateMessage
    ) -> Awaitable[None]:
        # Even though we're an 'async' handler, we want to run this
        # synchronously to make sure the create happens immediately.
        _impl1.handle_account_channel_create_message(self, msg)
        return self._await_none()

    @msg_in.handler
    def _handle_account_channel_destroy_message(
        self, msg: batocloud.AccountChannelDestroyMessage
    ) -> Awaitable[None]:
        # Even though we're an 'async' handler, we want to run this
        # synchronously to make sure the destroy happens immediately.
        _impl1.handle_account_channel_destroy_message(self, msg)
        return self._await_none()

    async def _await_none(self) -> None:
        return None

    # Make sure we've defined handlers for everything in our protocol.
    msg_in.validate(log_only=True)
