"""Main entry point for basntask tasks."""

from __future__ import annotations

import os
import sys
import signal
from pathlib import Path
from typing import TYPE_CHECKING

from efro.log import setup_logging, LogLevel
from efro.error import CleanError
from efro.dataclassio import dataclass_from_json, dataclass_to_json
from basntask.taskenv import TaskEnv

if TYPE_CHECKING:
    from types import FrameType
    from signal import Signals

    from basntask.command import TaskCommand


def main() -> None:
    """Main entry point for servernode task processes."""

    from baserver.enums import ServerNodeRunMode

    from basntask.command import TaskCommand

    if len(sys.argv) != 2:
        raise RuntimeError('Expected a single payload arg')

    env = dataclass_from_json(TaskEnv, sys.argv[-1])

    # Wire up a handler so we can gracefully handle TERM signals.
    signal.signal(signal.SIGTERM, TaskCommand.handle_term_signal)

    # Dump our PID so basn can talk to us. Write a temp file and then
    # rename it into place to hopefully keep things atomic.
    with open('.basntask.pid.tmp', 'w', encoding='utf-8') as outfile:
        outfile.write(str(os.getpid()))
    os.rename('.basntask.pid.tmp', '.basntask.pid')

    # Set up logging to our structured file.
    setup_logging(
        'basntask.log',
        # Currently we just globally set INFO or DEBUG depending on
        # basn's run mode, but we probably should allow the Command to
        # override this with a more nuanced setup.
        (
            LogLevel.DEBUG
            if env.run_mode.level >= ServerNodeRunMode.DEBUG.level
            else LogLevel.INFO
        ),
        # Ship stdout/stderr to our structured log too.
        log_stdout_stderr=True,
        # But not other way around (for efficiency).
        echo_to_stderr=False,
    )

    # We were probably run with PYTHONPATH set to help us find our
    # common packages (efro, etc). However, our payload might include
    # different versions of those same packages, so let's make sure
    # PYTHONPATH is unset before launching any child task so the child
    # picks up its own bundled stuff.
    if 'PYTHONPATH' in os.environ:
        del os.environ['PYTHONPATH']

    TaskCommand.set_env(env)

    result = env.command.run()

    # We get run detached so we don't have a parent process to catch our
    # results; let's write a simple 'success' file if we finish cleanly
    # so whoever is watching can know we succeeded if it is present and
    # can assume we failed otherwise.
    Path('.basntask.success').touch()

    # We run detached so we need to write our results to a file to
    # communicate them back to basn.
    with open('.basntask.result.tmp', 'w', encoding='utf-8') as outfile:
        outfile.write(dataclass_to_json(result))
    os.rename('.basntask.result.tmp', '.basntask.result')


if __name__ == '__main__':
    try:
        main()
    except CleanError as exc:
        exc.pretty_print()
