"""For testing various basn subtask mechanisms."""

from __future__ import annotations

import os
import time
import logging
from typing import TYPE_CHECKING

from dataclasses import dataclass

from typing_extensions import override
from efro.dataclassio import ioprepped
from baserver.cloudbuild import (
    RecipeTypeID,
    RecipeComponent,
    Recipe,
    CloudBuildContext,
)

from basntask.command import TaskCommand, TaskCommandTypeID
from basntask.result import TaskResult, TaskOutput, TaskOutputTypeID

if TYPE_CHECKING:
    pass


@ioprepped
@dataclass
class CloudBuildCommand(TaskCommand):
    """Run a cloud build."""

    recipe_type_id: RecipeTypeID
    input_data: RecipeComponent
    input_cloud_file_paths: dict[str, str]

    @override
    @classmethod
    def get_type_id(cls) -> TaskCommandTypeID:
        return TaskCommandTypeID.CLOUD_BUILD

    @override
    def run(self) -> TaskResult:

        logging.debug('Staging %s build...', self.recipe_type_id.value)
        starttime = time.monotonic()

        # Where we stick our output files.
        outdirname = 'outfiles'
        os.mkdir(outdirname)

        recipe_type = Recipe.get_type(self.recipe_type_id)
        context = CloudBuildContext(recipe_type, output_dir=outdirname)

        # Stage all input files on disk.
        stagecount = 0
        for cloudfileid in self.input_data.get_cloud_file_ids():
            assert cloudfileid in self.input_cloud_file_paths
            assert os.path.isfile(self.input_cloud_file_paths[cloudfileid])
            context.set_path_for_input_file(
                self.input_cloud_file_paths[cloudfileid], cloudfileid
            )
            stagecount += 1
        logging.debug(
            'Staged %s input files in %.3fs.',
            stagecount,
            time.monotonic() - starttime,
        )

        # Run the build.
        execstarttime = time.monotonic()
        logging.debug('Execing %s build...', self.recipe_type_id.value)
        output_data = context.exec_build(self.input_data)
        logging.debug(
            'Exec completed in %.3fs.', time.monotonic() - execstarttime
        )

        output_files_as_paths = context.get_output_files_as_paths()
        return TaskResult(
            output=CloudBuildTaskOutput(output_data, output_files_as_paths)
        )


@ioprepped
@dataclass
class CloudBuildTaskOutput(TaskOutput):
    """Output for a cloud-build-task."""

    output_data: RecipeComponent | None
    output_cloud_file_paths: dict[str, str]

    @override
    @classmethod
    def get_type_id(cls) -> TaskOutputTypeID:
        return TaskOutputTypeID.CLOUD_BUILD
